# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Dump the list of all dead urls in a `.txt` file.

Instructions:

```sh
python tensorflow_datasets/scripts/dead_url_checker.py
```

"""
from concurrent import futures

from absl import app
import requests

import tensorflow_datasets as tfds


def _get_status_code(url: str) -> int:
  try:
    return requests.head(url, timeout=10).status_code
  except (requests.exceptions.Timeout, requests.exceptions.ConnectionError):
    return -1


def main(_):
  # Legacy datasets
  urls = set(tfds.core.download.checksums.get_all_url_infos().keys())

  # Dataset-as-folder datasets
  # Could keep track of the dataset name, so the report clearly indicates which
  # dataset should be updated.
  url_infos = {
      name: tfds.builder_cls(name).url_infos
      for name in tfds.list_builders(with_community_datasets=False)
  }
  for url_info in url_infos.values():
    if url_info:
      urls |= url_info.keys()

  urls = sorted(urls)

  with futures.ThreadPoolExecutor(max_workers=100) as executor:
    all_codes = executor.map(_get_status_code, urls)

  print('\n************ Summary ************\n')
  total_errors = 0
  for url, code in zip(urls, all_codes):
    if code == requests.codes.ok:
      continue
    total_errors += 1
    print(f'{url} - status code: {code}')
  print(f'{total_errors} URLs had issues')


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""web_graph dataset."""

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.structured.web_graph import web_graph


class WebGraphTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for web_graph dataset."""

  DATASET_CLASS = web_graph.WebGraph
  SPLITS = {
      'train': 10,  # Number of fake train example
      'train_t': 10,  # Number of fake train example
      'test': 10,  # Number of fake train example
  }
  OVERLAPPING_SPLITS = ['train', 'train_t']

  SKIP_TF1_GRAPH_MODE = True

  # The different configs are only different in the name of the files.
  BUILDER_CONFIG_NAMES_TO_TEST = ['in-dense']

  @classmethod
  def setUpClass(cls):
    web_graph.WebGraph.WEB_GRAPH_HOMEPAGE = cls.dummy_data
    web_graph.WebGraph.SHARDS = 1
    super().setUpClass()


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: skip-file

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: waymo_dataset.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x13waymo_dataset.proto\x12\x12waymo.open_dataset"\x8f\x06\n\x05Label\x12*\n\x03\x62ox\x18\x01'
    b' \x01(\x0b\x32\x1d.waymo.open_dataset.Label.Box\x12\x34\n\x08metadata\x18\x02'
    b' \x01(\x0b\x32".waymo.open_dataset.Label.Metadata\x12,\n\x04type\x18\x03'
    b' \x01(\x0e\x32\x1e.waymo.open_dataset.Label.Type\x12\n\n\x02id\x18\x04'
    b' \x01(\t\x12M\n\x1a\x64\x65tection_difficulty_level\x18\x05'
    b' \x01(\x0e\x32).waymo.open_dataset.Label.DifficultyLevel\x12L\n\x19tracking_difficulty_level\x18\x06'
    b' \x01(\x0e\x32).waymo.open_dataset.Label.DifficultyLevel\x12\x1f\n\x17num_lidar_points_in_box\x18\x07'
    b' \x01(\x05\x1a\xbf\x01\n\x03\x42ox\x12\x10\n\x08\x63\x65nter_x\x18\x01'
    b' \x01(\x01\x12\x10\n\x08\x63\x65nter_y\x18\x02'
    b' \x01(\x01\x12\x10\n\x08\x63\x65nter_z\x18\x03'
    b' \x01(\x01\x12\x0e\n\x06length\x18\x05 \x01(\x01\x12\r\n\x05width\x18\x04'
    b' \x01(\x01\x12\x0e\n\x06height\x18\x06'
    b' \x01(\x01\x12\x0f\n\x07heading\x18\x07'
    b' \x01(\x01"B\n\x04Type\x12\x10\n\x0cTYPE_UNKNOWN\x10\x00\x12\x0b\n\x07TYPE_3D\x10\x01\x12\x0b\n\x07TYPE_2D\x10\x02\x12\x0e\n\nTYPE_AA_2D\x10\x03\x1aN\n\x08Metadata\x12\x0f\n\x07speed_x\x18\x01'
    b' \x01(\x01\x12\x0f\n\x07speed_y\x18\x02'
    b' \x01(\x01\x12\x0f\n\x07\x61\x63\x63\x65l_x\x18\x03'
    b' \x01(\x01\x12\x0f\n\x07\x61\x63\x63\x65l_y\x18\x04'
    b' \x01(\x01"`\n\x04Type\x12\x10\n\x0cTYPE_UNKNOWN\x10\x00\x12\x10\n\x0cTYPE_VEHICLE\x10\x01\x12\x13\n\x0fTYPE_PEDESTRIAN\x10\x02\x12\r\n\tTYPE_SIGN\x10\x03\x12\x10\n\x0cTYPE_CYCLIST\x10\x04"8\n\x0f\x44ifficultyLevel\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07LEVEL_1\x10\x01\x12\x0b\n\x07LEVEL_2\x10\x02"2\n\x0ePolygon2dProto\x12\t\n\x01x\x18\x01'
    b' \x03(\x01\x12\t\n\x01y\x18\x02 \x03(\x01\x12\n\n\x02id\x18\x03'
    b' \x01(\t"\x1b\n\x0bMatrixShape\x12\x0c\n\x04\x64ims\x18\x01'
    b' \x03(\x05"O\n\x0bMatrixFloat\x12\x10\n\x04\x64\x61ta\x18\x01'
    b' \x03(\x02\x42\x02\x10\x01\x12.\n\x05shape\x18\x02'
    b' \x01(\x0b\x32\x1f.waymo.open_dataset.MatrixShape"O\n\x0bMatrixInt32\x12\x10\n\x04\x64\x61ta\x18\x01'
    b' \x03(\x05\x42\x02\x10\x01\x12.\n\x05shape\x18\x02'
    b' \x01(\x0b\x32\x1f.waymo.open_dataset.MatrixShape"l\n\nCameraName"^\n\x04Name\x12\x0b\n\x07UNKNOWN\x10\x00\x12\t\n\x05\x46RONT\x10\x01\x12\x0e\n\nFRONT_LEFT\x10\x02\x12\x0f\n\x0b\x46RONT_RIGHT\x10\x03\x12\r\n\tSIDE_LEFT\x10\x04\x12\x0e\n\nSIDE_RIGHT\x10\x05"]\n\tLaserName"P\n\x04Name\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x07\n\x03TOP\x10\x01\x12\t\n\x05\x46RONT\x10\x02\x12\r\n\tSIDE_LEFT\x10\x03\x12\x0e\n\nSIDE_RIGHT\x10\x04\x12\x08\n\x04REAR\x10\x05"\x1e\n\tTransform\x12\x11\n\ttransform\x18\x01'
    b' \x03(\x01"X\n\x08Velocity\x12\x0b\n\x03v_x\x18\x01'
    b' \x01(\x02\x12\x0b\n\x03v_y\x18\x02 \x01(\x02\x12\x0b\n\x03v_z\x18\x03'
    b' \x01(\x02\x12\x0b\n\x03w_x\x18\x04 \x01(\x01\x12\x0b\n\x03w_y\x18\x05'
    b' \x01(\x01\x12\x0b\n\x03w_z\x18\x06'
    b' \x01(\x01"\xa3\x03\n\x11\x43\x61meraCalibration\x12\x31\n\x04name\x18\x01'
    b' \x01(\x0e\x32#.waymo.open_dataset.CameraName.Name\x12\x11\n\tintrinsic\x18\x02'
    b' \x03(\x01\x12\x30\n\textrinsic\x18\x03'
    b' \x01(\x0b\x32\x1d.waymo.open_dataset.Transform\x12\r\n\x05width\x18\x04'
    b' \x01(\x05\x12\x0e\n\x06height\x18\x05'
    b' \x01(\x05\x12g\n\x19rolling_shutter_direction\x18\x06'
    b' \x01(\x0e\x32\x44.waymo.open_dataset.CameraCalibration.RollingShutterReadOutDirection"\x8d\x01\n\x1eRollingShutterReadOutDirection\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x11\n\rTOP_TO_BOTTOM\x10\x01\x12\x11\n\rLEFT_TO_RIGHT\x10\x02\x12\x11\n\rBOTTOM_TO_TOP\x10\x03\x12\x11\n\rRIGHT_TO_LEFT\x10\x04\x12\x12\n\x0eGLOBAL_SHUTTER\x10\x05"\xcd\x01\n\x10LaserCalibration\x12\x30\n\x04name\x18\x01'
    b' \x01(\x0e\x32".waymo.open_dataset.LaserName.Name\x12\x19\n\x11\x62\x65\x61m_inclinations\x18\x02'
    b' \x03(\x01\x12\x1c\n\x14\x62\x65\x61m_inclination_min\x18\x03'
    b' \x01(\x01\x12\x1c\n\x14\x62\x65\x61m_inclination_max\x18\x04'
    b' \x01(\x01\x12\x30\n\textrinsic\x18\x05'
    b' \x01(\x0b\x32\x1d.waymo.open_dataset.Transform"\xf6\x03\n\x07\x43ontext\x12\x0c\n\x04name\x18\x01'
    b' \x01(\t\x12\x42\n\x13\x63\x61mera_calibrations\x18\x02'
    b' \x03(\x0b\x32%.waymo.open_dataset.CameraCalibration\x12@\n\x12laser_calibrations\x18\x03'
    b' \x03(\x0b\x32$.waymo.open_dataset.LaserCalibration\x12\x30\n\x05stats\x18\x04'
    b' \x01(\x0b\x32!.waymo.open_dataset.Context.Stats\x1a\xa4\x02\n\x05Stats\x12J\n\x13laser_object_counts\x18\x01'
    b' \x03(\x0b\x32-.waymo.open_dataset.Context.Stats.ObjectCount\x12K\n\x14\x63\x61mera_object_counts\x18\x05'
    b' \x03(\x0b\x32-.waymo.open_dataset.Context.Stats.ObjectCount\x12\x13\n\x0btime_of_day\x18\x02'
    b' \x01(\t\x12\x10\n\x08location\x18\x03'
    b' \x01(\t\x12\x0f\n\x07weather\x18\x04'
    b' \x01(\t\x1aJ\n\x0bObjectCount\x12,\n\x04type\x18\x01'
    b' \x01(\x0e\x32\x1e.waymo.open_dataset.Label.Type\x12\r\n\x05\x63ount\x18\x02'
    b' \x01(\x05"\xb1\x01\n\nRangeImage\x12\x1e\n\x16range_image_compressed\x18\x02'
    b' \x01(\x0c\x12$\n\x1c\x63\x61mera_projection_compressed\x18\x03'
    b' \x01(\x0c\x12#\n\x1brange_image_pose_compressed\x18\x04'
    b' \x01(\x0c\x12\x38\n\x0brange_image\x18\x01'
    b' \x01(\x0b\x32\x1f.waymo.open_dataset.MatrixFloatB\x02\x18\x01"\x94\x02\n\x0b\x43\x61meraImage\x12\x31\n\x04name\x18\x01'
    b' \x01(\x0e\x32#.waymo.open_dataset.CameraName.Name\x12\r\n\x05image\x18\x02'
    b' \x01(\x0c\x12+\n\x04pose\x18\x03'
    b' \x01(\x0b\x32\x1d.waymo.open_dataset.Transform\x12.\n\x08velocity\x18\x04'
    b' \x01(\x0b\x32\x1c.waymo.open_dataset.Velocity\x12\x16\n\x0epose_timestamp\x18\x05'
    b' \x01(\x01\x12\x0f\n\x07shutter\x18\x06'
    b' \x01(\x01\x12\x1b\n\x13\x63\x61mera_trigger_time\x18\x07 \x01(\x01\x12'
    b' \n\x18\x63\x61mera_readout_done_time\x18\x08'
    b' \x01(\x01"l\n\x0c\x43\x61meraLabels\x12\x31\n\x04name\x18\x01'
    b' \x01(\x0e\x32#.waymo.open_dataset.CameraName.Name\x12)\n\x06labels\x18\x02'
    b' \x03(\x0b\x32\x19.waymo.open_dataset.Label"\xa1\x01\n\x05Laser\x12\x30\n\x04name\x18\x01'
    b' \x01(\x0e\x32".waymo.open_dataset.LaserName.Name\x12\x32\n\nri_return1\x18\x02'
    b' \x01(\x0b\x32\x1e.waymo.open_dataset.RangeImage\x12\x32\n\nri_return2\x18\x03'
    b' \x01(\x0b\x32\x1e.waymo.open_dataset.RangeImage"\xc0\x03\n\x05\x46rame\x12,\n\x07\x63ontext\x18\x01'
    b' \x01(\x0b\x32\x1b.waymo.open_dataset.Context\x12\x18\n\x10timestamp_micros\x18\x02'
    b' \x01(\x03\x12+\n\x04pose\x18\x03'
    b' \x01(\x0b\x32\x1d.waymo.open_dataset.Transform\x12/\n\x06images\x18\x04'
    b' \x03(\x0b\x32\x1f.waymo.open_dataset.CameraImage\x12)\n\x06lasers\x18\x05'
    b' \x03(\x0b\x32\x19.waymo.open_dataset.Laser\x12/\n\x0claser_labels\x18\x06'
    b' \x03(\x0b\x32\x19.waymo.open_dataset.Label\x12@\n\x16projected_lidar_labels\x18\t'
    b' \x03(\x0b\x32'
    b' .waymo.open_dataset.CameraLabels\x12\x37\n\rcamera_labels\x18\x08'
    b' \x03(\x0b\x32'
    b' .waymo.open_dataset.CameraLabels\x12:\n\x0eno_label_zones\x18\x07'
    b' \x03(\x0b\x32".waymo.open_dataset.Polygon2dProto'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, 'waymo_dataset_pb2', globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _MATRIXFLOAT.fields_by_name['data']._options = None
  _MATRIXFLOAT.fields_by_name['data']._serialized_options = b'\020\001'
  _MATRIXINT32.fields_by_name['data']._options = None
  _MATRIXINT32.fields_by_name['data']._serialized_options = b'\020\001'
  _RANGEIMAGE.fields_by_name['range_image']._options = None
  _RANGEIMAGE.fields_by_name['range_image']._serialized_options = b'\030\001'
  _LABEL._serialized_start = 44
  _LABEL._serialized_end = 827
  _LABEL_BOX._serialized_start = 400
  _LABEL_BOX._serialized_end = 591
  _LABEL_BOX_TYPE._serialized_start = 525
  _LABEL_BOX_TYPE._serialized_end = 591
  _LABEL_METADATA._serialized_start = 593
  _LABEL_METADATA._serialized_end = 671
  _LABEL_TYPE._serialized_start = 673
  _LABEL_TYPE._serialized_end = 769
  _LABEL_DIFFICULTYLEVEL._serialized_start = 771
  _LABEL_DIFFICULTYLEVEL._serialized_end = 827
  _POLYGON2DPROTO._serialized_start = 829
  _POLYGON2DPROTO._serialized_end = 879
  _MATRIXSHAPE._serialized_start = 881
  _MATRIXSHAPE._serialized_end = 908
  _MATRIXFLOAT._serialized_start = 910
  _MATRIXFLOAT._serialized_end = 989
  _MATRIXINT32._serialized_start = 991
  _MATRIXINT32._serialized_end = 1070
  _CAMERANAME._serialized_start = 1072
  _CAMERANAME._serialized_end = 1180
  _CAMERANAME_NAME._serialized_start = 1086
  _CAMERANAME_NAME._serialized_end = 1180
  _LASERNAME._serialized_start = 1182
  _LASERNAME._serialized_end = 1275
  _LASERNAME_NAME._serialized_start = 1195
  _LASERNAME_NAME._serialized_end = 1275
  _TRANSFORM._serialized_start = 1277
  _TRANSFORM._serialized_end = 1307
  _VELOCITY._serialized_start = 1309
  _VELOCITY._serialized_end = 1397
  _CAMERACALIBRATION._serialized_start = 1400
  _CAMERACALIBRATION._serialized_end = 1819
  _CAMERACALIBRATION_ROLLINGSHUTTERREADOUTDIRECTION._serialized_start = 1678
  _CAMERACALIBRATION_ROLLINGSHUTTERREADOUTDIRECTION._serialized_end = 1819
  _LASERCALIBRATION._serialized_start = 1822
  _LASERCALIBRATION._serialized_end = 2027
  _CONTEXT._serialized_start = 2030
  _CONTEXT._serialized_end = 2532
  _CONTEXT_STATS._serialized_start = 2240
  _CONTEXT_STATS._serialized_end = 2532
  _CONTEXT_STATS_OBJECTCOUNT._serialized_start = 2458
  _CONTEXT_STATS_OBJECTCOUNT._serialized_end = 2532
  _RANGEIMAGE._serialized_start = 2535
  _RANGEIMAGE._serialized_end = 2712
  _CAMERAIMAGE._serialized_start = 2715
  _CAMERAIMAGE._serialized_end = 2991
  _CAMERALABELS._serialized_start = 2993
  _CAMERALABELS._serialized_end = 3101
  _LASER._serialized_start = 3104
  _LASER._serialized_end = 3265
  _FRAME._serialized_start = 3268
  _FRAME._serialized_end = 3716
# @@protoc_insertion_point(module_scope)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""wiki_table_text dataset."""

from etils import epath
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.structured.wiki_table_text import wiki_table_text


class WikiTableTextTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for wiki_table_text dataset."""

  DATASET_CLASS = wiki_table_text.WikiTableText
  SPLITS = {
      'train': 4,
      'validation': 2,
      'test': 3,
  }

  DL_EXTRACT_RESULT = {
      'train_path': 'MSRA_NLC.Table2Text.train',
      'dev_path': 'MSRA_NLC.Table2Text.dev',
      'test_path': 'MSRA_NLC.Table2Text.test',
  }

  def test_generate_examples(self):
    expected_examples = [
        {
            'input_text': {
                'table': [
                    {
                        'column_header': 'subj_title_',
                        'row_number': 1,
                        'content': 'central bank of russia',
                    },
                    {
                        'column_header': 'subj_subtitle',
                        'row_number': 1,
                        'content': 'chairmen',
                    },
                    {
                        'column_header': 'name',
                        'row_number': 1,
                        'content': 'viktor gerashchenko',
                    },
                    {
                        'column_header': 'year',
                        'row_number': 1,
                        'content': '1992—1994',
                    },
                ]
            },
            'target_text': (
                'viktor gerashchenko is the chairman of central bank of russia '
                'during 1992 to 1994'
            ),
        },
        {
            'input_text': {
                'table': [
                    {
                        'column_header': 'subj_title_',
                        'row_number': 1,
                        'content': 'ace frehley',
                    },
                    {
                        'column_header': 'subj_subtitle',
                        'row_number': 1,
                        'content': 'discography',
                    },
                    {
                        'column_header': 'year',
                        'row_number': 1,
                        'content': '2006',
                    },
                    {
                        'column_header': 'title',
                        'row_number': 1,
                        'content': 'greatest hits live',
                    },
                ]
            },
            'target_text': (
                'greatest hits live is a discography of ace frehlely in 2006'
            ),
        },
    ]
    dataset = wiki_table_text.WikiTableText()

    def mock_open(*_):
      current_path = epath.Path(__file__).parent
      return open(current_path / 'test_data/examples.csv')

    with epath.testing.mock_epath(open=mock_open):
      examples = list(dataset._generate_examples(''))
      for i, (_, example) in enumerate(examples):
        self.assertCountEqual(example, expected_examples[i])
      assert len(examples) == len(expected_examples)


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for librispeech dataset module."""

from tensorflow_datasets import testing
from tensorflow_datasets.audio.userlibri_audio_data import userlibri_audio_data


class UserLibriAudioTest(testing.DatasetBuilderTestCase):
  DATASET_CLASS = userlibri_audio_data.UserLibriAudio
  SPLITS = {
      # List number of fake train examples.
      "test-clean_speaker-121-book-1041": 2,
      "test-clean_speaker-121-book-1989": 2,
      "test-other_speaker-3005-book-76": 2,
      "test-other_speaker-8461-book-6328": 1,
      "test-other_speaker-8461-book-9189": 1,
  }
  DL_DOWNLOAD_RESULT = ""
  SKIP_CHECKSUMS = True


if __name__ == "__main__":
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.features.audio_feature."""

import array
import pathlib
import tempfile

from absl.testing import parameterized
import numpy as np
import pydub
import tensorflow as tf
from tensorflow_datasets import testing
from tensorflow_datasets.core import features


class AudioFeatureTest(
    testing.FeatureExpectationsTestCase, parameterized.TestCase
):

  @parameterized.product(dtype=[np.int64, tf.int64], num_channels=[1, 2, 8])
  def test_numpy_array(self, dtype, num_channels):
    np_audio = _create_np_audio(num_channels, np.int64)

    self.assertFeature(
        feature=features.Audio(
            sample_rate=1000, shape=_shape_for_channels(num_channels)
        ),
        shape=_shape_for_channels(num_channels),
        dtype=dtype,
        tests=[
            testing.FeatureExpectationItem(
                value=np_audio,
                expected=np_audio,
            ),
        ],
        test_attributes=dict(
            _file_format=None,
            sample_rate=1000,
        ),
    )

  @parameterized.product(dtype=[np.float32, tf.float32], num_channels=[1, 2, 8])
  def test_numpy_array_float(self, dtype, num_channels):
    np_audio = _create_np_audio(num_channels, np.float32)

    self.assertFeature(
        feature=features.Audio(
            dtype=dtype, shape=_shape_for_channels(num_channels)
        ),
        shape=_shape_for_channels(num_channels),
        dtype=np.float32,
        tests=[
            testing.FeatureExpectationItem(
                value=np_audio,
                expected=np_audio,
            ),
        ],
    )

  @parameterized.product(
      dtype=[np.int16, np.int32],
      lazy_decode=[True, False],
      num_channels=[1, 2, 8],
  )
  def test_wav_file(self, dtype, lazy_decode, num_channels):
    file_format = 'wav'

    np_audio = _create_np_audio(num_channels, dtype)
    _, tmp_file = tempfile.mkstemp()
    _write_audio_file(np_audio, tmp_file, file_format)

    self.assertFeature(
        feature=features.Audio(
            file_format=file_format,
            shape=_shape_for_channels(num_channels),
            dtype=dtype,
            lazy_decode=lazy_decode,
        ),
        shape=_shape_for_channels(num_channels),
        dtype=dtype,
        tests=[
            testing.FeatureExpectationItem(
                value=tmp_file,
                expected=np_audio,
            ),
            testing.FeatureExpectationItem(
                value=pathlib.Path(tmp_file),
                expected=np_audio,
            ),
        ],
        test_attributes=dict(
            _file_format=file_format,
        ),
    )

  @parameterized.product(lazy_decode=[True, False], num_channels=[1, 2, 8])
  def test_flac_file(self, lazy_decode, num_channels):
    if lazy_decode:
      if 'dev' in tf.__version__:
        self.skipTest('`tensorflow_io` is not compatible with `tf-nightly`')

    dtype = np.int16
    file_format = 'flac'

    np_audio = _create_np_audio(num_channels, dtype)
    _, tmp_file = tempfile.mkstemp()
    _write_audio_file(np_audio, tmp_file, file_format)

    self.assertFeature(
        feature=features.Audio(
            file_format=file_format,
            shape=_shape_for_channels(num_channels),
            dtype=dtype,
            lazy_decode=lazy_decode,
        ),
        shape=_shape_for_channels(num_channels),
        dtype=dtype,
        tests=[
            testing.FeatureExpectationItem(
                value=tmp_file,
                expected=np_audio,
            ),
            testing.FeatureExpectationItem(
                value=pathlib.Path(tmp_file),
                expected=np_audio,
            ),
        ],
        test_attributes=dict(
            _file_format=file_format,
        ),
    )

  @parameterized.product(
      dtype=[np.int16, np.int32],
      lazy_decode=[True, False],
      num_channels=[1, 2, 8],
  )
  def test_file_object(self, dtype, lazy_decode, num_channels):
    file_format = 'wav'

    np_audio = _create_np_audio(num_channels, dtype)
    _, tmp_file = tempfile.mkstemp()
    _write_audio_file(np_audio, tmp_file, file_format)

    class GFileWithSeekOnRead(tf.io.gfile.GFile):
      """Wrapper around GFile which is reusable across multiple read() calls.

      This is needed because assertFeature reuses the same
      FeatureExpectationItem several times.
      """

      def read(self, *args, **kwargs):
        data_read = super(GFileWithSeekOnRead, self).read(*args, **kwargs)
        self.seek(0)
        return data_read

    with GFileWithSeekOnRead(tmp_file, 'rb') as file_obj:
      self.assertFeature(
          feature=features.Audio(
              file_format=file_format,
              shape=_shape_for_channels(num_channels),
              dtype=dtype,
              lazy_decode=lazy_decode,
          ),
          shape=_shape_for_channels(num_channels),
          dtype=dtype,
          tests=[
              testing.FeatureExpectationItem(
                  value=file_obj,
                  expected=np_audio,
              ),
          ],
      )

  def test_sample_rate_property(self):
    self.assertEqual(features.Audio(sample_rate=1600).sample_rate, 1600)

  def test_lazy_decode_error_file(self):
    with self.assertRaisesWithPredicateMatch(
        ValueError, 'Acceptable `dtype` for lazy loading'
    ):
      features.Audio(file_format='flac', lazy_decode=True)

  @parameterized.parameters([(None,), 2, (10,)], [(None, 2), 1, (10, 2)])
  def test_number_of_channels_is_changed(
      self, feature_shape, example_num_channels, example_output_shape
  ):
    feature = features.Audio(
        file_format='wav', dtype=np.int16, shape=feature_shape
    )
    _, tmp_file = tempfile.mkstemp()
    np_audio = _create_np_audio(example_num_channels, np.int16)
    _write_audio_file(np_audio, tmp_file, 'wav')
    self.assertEqual(
        feature.encode_example(tmp_file).shape, example_output_shape
    )


def _shape_for_channels(num_channels, *, length=None):
  """Returns the shape."""
  return (length,) if num_channels == 1 else (length, num_channels)


def _create_np_audio(num_channels: int, dtype: np.dtype) -> np.ndarray:
  """Creates a random audio `np.array`."""
  if np.issubdtype(dtype, np.integer):
    dtype_info = np.iinfo(dtype)
  else:
    dtype_info = np.finfo(dtype)

  np_audio = np.random.randint(
      max(-(2**10), int(dtype_info.min)),
      min(2**10, int(dtype_info.max)),
      size=_shape_for_channels(num_channels, length=10),
  )

  return np_audio.astype(dtype)


def _write_audio_file(np_audio, path, file_format):
  """Creates a random audio file with the given file format."""
  dtype = np_audio.dtype

  if (file_format == 'flac' and dtype not in [np.int16]) or (
      file_format == 'wav' and dtype not in [np.int16, np.int32]
  ):
    raise ValueError(
        f'Can\'t save audio in "{file_format.upper()}" format with dtype '
        f'`{np_audio.dtype}`.'
    )

  data = array.array(np.sctype2char(dtype), np_audio.reshape((-1,)))
  sample_width = np.dtype(dtype).alignment
  num_channels = np_audio.shape[1] if len(np_audio.shape) == 2 else 1

  pydub.AudioSegment(
      data=data,
      sample_width=sample_width,
      channels=num_channels,
      frame_rate=1,
  ).export(path, format=file_format)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Load Datasets without reading dataset generation code."""

import os
import typing
from typing import Any, List, Optional, Tuple, Type

from etils import epath
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import dataset_info
from tensorflow_datasets.core import logging as tfds_logging
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import registered
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.features import feature as feature_lib
from tensorflow_datasets.core.proto import dataset_info_pb2
from tensorflow_datasets.core.utils import error_utils
from tensorflow_datasets.core.utils import file_utils
from tensorflow_datasets.core.utils import version as version_lib
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf


class ReadOnlyBuilder(
    dataset_builder.FileReaderBuilder, skip_registration=True
):
  """Generic DatasetBuilder loading from a directory."""

  @tfds_logging.builder_init()
  def __init__(
      self,
      builder_dir: epath.PathLike,
      *,
      info_proto: Optional[dataset_info_pb2.DatasetInfo] = None,
  ):
    """Constructor.

    Args:
      builder_dir: Directory of the dataset to load (e.g.
        `~/tensorflow_datasets/mnist/3.0.0/`)
      info_proto: DatasetInfo describing the name, config, etc of the requested
        dataset. Note that this overwrites dataset info that may be present in
        builder_dir.

    Raises:
      FileNotFoundError: If the builder_dir does not exist.
    """
    builder_dir = os.path.expanduser(builder_dir)
    if not info_proto:
      info_proto = dataset_info.read_proto_from_builder_dir(builder_dir)
    self._info_proto = info_proto

    self.name = info_proto.name
    self.VERSION = version_lib.Version(info_proto.version)  # pylint: disable=invalid-name
    self.RELEASE_NOTES = info_proto.release_notes or {}  # pylint: disable=invalid-name

    if info_proto.module_name:
      # Overwrite the module so documenting `ReadOnlyBuilder` point to the
      # original source code.
      self.__module__ = info_proto.module_name

    builder_config = dataset_builder.BuilderConfig.from_dataset_info(info_proto)
    # __init__ will call _build_data_dir, _create_builder_config,
    # _pick_version to set the data_dir, config, and version
    super().__init__(
        data_dir=builder_dir,
        config=builder_config,
        version=info_proto.version,
    )

    # For pickling, should come after super.__init__ which is setting that same
    # _original_state attribute.
    self._original_state = dict(builder_dir=builder_dir)

    if self.info.features is None:
      raise ValueError(
          f'Cannot restore {self.info.full_name}. It likely means the dataset '
          'was generated with an old TFDS version (<=3.2.1).'
      )

  def _create_builder_config(
      self, builder_config: Optional[dataset_builder.BuilderConfig]
  ) -> Optional[dataset_builder.BuilderConfig]:
    return builder_config  # BuilderConfig is created in __init__

  def _pick_version(self, version: str) -> utils.Version:
    return utils.Version(version)

  def _build_data_dir(self, data_dir: str) -> Tuple[str, str]:
    return data_dir, data_dir  # _data_dir_root, _data_dir are builder_dir.

  def _info(self) -> dataset_info.DatasetInfo:
    return dataset_info.DatasetInfo.from_proto(self, self._info_proto)

  def _download_and_prepare(self, **kwargs):  # pylint: disable=arguments-differ
    # DatasetBuilder.download_and_prepare is a no-op as self.data_dir already
    # exists.
    raise AssertionError("ReadOnlyBuilder can't be generated.")


def builder_from_directory(
    builder_dir: epath.PathLike,
) -> dataset_builder.DatasetBuilder:
  """Loads a `tfds.core.DatasetBuilder` from the given generated dataset path.

  Reconstructs the `tfds.core.DatasetBuilder` without requiring the original
  generation code.

  From `<builder_dir>/features.json` it infers the structure (feature names,
  nested dict,...) and content (image, sequence,...) of the dataset. The
  serialization format is defined in
  `tfds.features.FeatureConnector` in `to_json()`.

  Note: This function only works for datasets generated with TFDS `4.0.0` or
  above.

  Args:
    builder_dir: `str`, path of the directory containing the dataset to read (
      e.g. `~/tensorflow_datasets/mnist/3.0.0/`).

  Returns:
    builder: `tfds.core.DatasetBuilder`, builder for dataset at the given path.
  """
  return ReadOnlyBuilder(builder_dir=builder_dir)


def builder_from_directories(
    builder_dirs: List[epath.PathLike],
    *,
    filetype_suffix: Optional[str] = None,  # DEPRECATED
) -> dataset_builder.DatasetBuilder:
  """Loads a `tfds.core.DatasetBuilder` from the given generated dataset path.

  When a dataset is spread out over multiple folders, then this function can be
  used to easily read from all builder dirs.

  Note that the data in each folder must have the same features, dataset name,
  and version.

  Some examples of when a dataset might be spread out over multiple folders:

  - in reinforcement learning, multiple agents each produce a dataset
  - each day a new dataset is produced based on new incoming events

  Arguments:
    builder_dirs: the list of builder dirs from which the data should be read.
    filetype_suffix: DEPRECATED PLEASE DO NOT USE. The filetype suffix (e.g.
      'tfrecord') that is used if the file format is not specified in the
      DatasetInfo.

  Returns:
    the read only dataset builder that is configured to read from all the given
    builder dirs.
  """
  if not builder_dirs:
    raise ValueError('No builder dirs were given!')

  dataset_infos = utils.tree.parallel_map(
      dataset_info.read_proto_from_builder_dir, {d: d for d in builder_dirs}
  )

  # TODO(tfds): assert that the features of all datasets are identical.

  def get_split_dict(builder_dir, dataset_info_proto) -> splits_lib.SplitDict:
    filename_template = naming.ShardedFileTemplate(
        dataset_name=dataset_info_proto.name,
        data_dir=builder_dir,
        filetype_suffix=dataset_info_proto.file_format or filetype_suffix,
    )
    return splits_lib.SplitDict.from_proto(
        repeated_split_infos=dataset_info_proto.splits,
        filename_template=filename_template,
    )

  merged_split_dict = splits_lib.SplitDict.merge_multiple(
      [
          get_split_dict(builder_dir, dataset_info_proto)
          for builder_dir, dataset_info_proto in dataset_infos.items()
      ]
  )

  # We create the ReadOnlyBuilder for a random builder_dir and then update the
  # splits to capture the splits from all builder dirs.
  random_builder_dir = builder_dirs[0]
  random_builder = ReadOnlyBuilder(
      builder_dir=random_builder_dir,
      info_proto=dataset_infos[random_builder_dir],
  )
  random_builder.info.set_splits(merged_split_dict)
  return random_builder


def builder_from_metadata(
    builder_dir: epath.PathLike,
    info_proto: dataset_info_pb2.DatasetInfo,
) -> dataset_builder.DatasetBuilder:
  """Loads a `tfds.core.DatasetBuilder` from the given metadata.

  Reconstructs the `tfds.core.DatasetBuilder` without requiring the original
  generation code. The given `info_proto` overrides whatever dataset info is in
  `builder_dir`.

  The dataset structure (feature names, nested dict,...) and content (image,
  sequence, ...) is used from the given DatasetInfo.

  Args:
    builder_dir: path of the directory containing the dataset to read (e.g.
      `~/tensorflow_datasets/mnist/3.0.0/`). Dataset info in this folder is
      overridden by the `info_proto`.
    info_proto: DatasetInfo describing the name, config, features, etc of the
      requested dataset.

  Returns:
    builder: `tfds.core.DatasetBuilder`, builder for dataset at the given path.
  """
  return ReadOnlyBuilder(
      builder_dir=builder_dir,
      info_proto=info_proto,
  )


@error_utils.reraise_with_context(registered.DatasetNotFoundError)
def builder_from_files(
    name: str,
    **builder_kwargs: Any,
) -> dataset_builder.DatasetBuilder:
  """Loads a `tfds.core.DatasetBuilder` from files, auto-infering location.

  This function is similar to `tfds.builder` (same signature), but creates
  the `tfds.core.DatasetBuilder` directly from files, without loading
  original generation source code.

  It does not support:

   * namespaces (e.g. 'kaggle:dataset')
   * config objects (`dataset/config` valid, but not `config=MyConfig()`)
   * `version='experimental_latest'`

  Args:
    name: Dataset name.
    **builder_kwargs: `tfds.core.DatasetBuilder` kwargs.

  Returns:
    builder: The loaded dataset builder.

  Raises:
    DatasetNotFoundError: If the dataset cannot be loaded.
  """
  # Find and load dataset builder.
  builder_dir = _find_builder_dir(name, **builder_kwargs)
  if builder_dir is None:
    data_dirs = file_utils.list_data_dirs(
        given_data_dir=builder_kwargs.get('data_dir')
    )
    raise registered.DatasetNotFoundError(
        f'Could not find dataset files for: {name}. Make sure the dataset '
        f'has been generated in: {data_dirs}. If the dataset has configs, you '
        'might have to specify the config name.'
    )
  return builder_from_directory(builder_dir)


def _find_builder_dir(name: str, **builder_kwargs: Any) -> Optional[str]:
  """Search whether the given dataset is present on disk and return its path.

  Note:

   * If the dataset is present, but is legacy (no feature config file), None
     is returned.
   * If the config isn't specified, the function tries to infer the default
     config name from the original `DatasetBuilder`.
   * The function searches in all `data_dir` registered with
     `tfds.core.add_data_dir`. If the dataset exists in multiple dirs, an error
     is raised.

  Args:
    name: Builder name (e.g. `my_ds`, `my_ds/config`, `my_ds:1.2.0`,...)
    **builder_kwargs: `tfds.core.DatasetBuilder` kwargs.

  Returns:
    path: The dataset path found, or None if the dataset isn't found.
  """
  # Normalize builder kwargs
  name, builder_kwargs = naming.parse_builder_name_kwargs(
      name, **builder_kwargs
  )
  version = builder_kwargs.pop('version', None)
  config = builder_kwargs.pop('config', None)
  data_dir = builder_kwargs.pop('data_dir', None)

  # Builder cannot be found if it uses:
  # * namespace
  # * version='experimental_latest'
  # * config objects (rather than `str`)
  # * custom DatasetBuilder.__init__ kwargs
  if (
      name.namespace
      or version == 'experimental_latest'
      or isinstance(config, dataset_builder.BuilderConfig)
      or builder_kwargs
  ):
    error_msgs = ['Builder cannot be loaded from files if it uses:']
    if name.namespace:
      error_msgs.append(f'* namespaces (here, {name.namespace} is used)')
    if version == 'experimental_latest':
      error_msgs.append('* `experimental_latest` as requested version.')
    if isinstance(config, dataset_builder.BuilderConfig):
      error_msgs.append('* config objects (rather than `str`).')
    if builder_kwargs:
      error_msgs.append('* custom DatasetBuilder.__init__ kwargs.')
    error_utils.add_context('\t'.join(error_msgs))
    return None

  # Search the dataset across all registered data_dirs
  all_builder_dirs = []
  all_data_dirs = file_utils.list_data_dirs(given_data_dir=data_dir)
  for current_data_dir in all_data_dirs:
    builder_dir = _find_builder_dir_single_dir(
        name.name,
        data_dir=current_data_dir,
        version_str=str(version) if version else None,
        config_name=config,
    )
    if builder_dir:
      all_builder_dirs.append(builder_dir)

  if not all_builder_dirs:
    all_dirs_str = '\n\t- '.join([''] + [str(dir) for dir in all_data_dirs])
    error_msg = f'No registered data_dirs were found in:{all_dirs_str}\n'
    error_utils.add_context(error_msg)
    return None

  elif len(all_builder_dirs) != 1:
    # Rather than raising error every time, we could potentially be smarter
    # and load the most recent version across all files, but should be
    # careful when partial version is requested ('my_dataset:3.*.*').
    # Could add some `MultiDataDirManager` API:
    # ```
    # manager = MultiDataDirManager(given_data_dir=data_dir)
    # with manager.merge_data_dirs() as virtual_data_dir:
    #  virtual_builder_dir = _find_builder_dir(name, data_dir=virtual_data_dir)
    #  builder_dir = manager.resolve(virtual_builder_dir)
    # ```
    raise ValueError(
        f'Dataset {name} detected in multiple locations: {all_builder_dirs}. '
        'Please resolve the ambiguity by explicitly setting `data_dir=`.'
    )

  return all_builder_dirs[0]


def _get_dataset_dir(
    builder_dir: epath.Path,
    *,
    version_str: str,
    config_name: Optional[str] = None,
) -> epath.Path:
  """Returns the path for the given dataset, config and version."""
  dataset_dir = builder_dir
  if config_name:
    dataset_dir = dataset_dir / config_name
  return dataset_dir / version_str


def _contains_dataset(dataset_dir: epath.PathLike) -> bool:
  try:
    return feature_lib.make_config_path(dataset_dir).exists()
  except (OSError, tf.errors.PermissionDeniedError):
    return False


def _find_builder_dir_single_dir(
    builder_name: str,
    *,
    data_dir: epath.PathLike,
    config_name: Optional[str] = None,
    version_str: Optional[str] = None,
) -> Optional[str]:
  """Same as `find_builder_dir` but requires explicit dir."""

  builder_dir = epath.Path(data_dir) / builder_name

  # If the version is specified, check if the dataset dir exists and return.
  if version_str and version_lib.Version.is_valid(version_str):
    dataset_dir = _get_dataset_dir(
        builder_dir=builder_dir,
        version_str=version_str,
        config_name=config_name,
    )
    if _contains_dataset(dataset_dir):
      return os.fspath(dataset_dir)

  # If no config_name or an empty string was given, we try to find the default
  # config and load the dataset for that.
  if not config_name:
    default_config_name = _get_default_config_name(
        builder_dir=builder_dir, name=builder_name
    )
    if default_config_name:
      return _find_builder_dir_single_dir(
          builder_name=builder_name,
          data_dir=data_dir,
          config_name=default_config_name,
          version_str=version_str,
      )

  # Dataset wasn't found, try to find a suitable available version.
  found_version_str = _get_version_str(
      builder_dir, config_name=config_name, requested_version=version_str
  )
  if found_version_str and (
      version_str is None or found_version_str != version_str
  ):
    return _find_builder_dir_single_dir(
        builder_name=builder_name,
        data_dir=data_dir,
        config_name=config_name,
        version_str=found_version_str,
    )

  # If no builder found, we populate the error_context with useful information
  # and return None.
  error_utils.add_context(
      'No builder could be found in the directory: '
      f'{data_dir} for the builder: {builder_name}.'
  )
  return None


def _get_default_config_name(
    builder_dir: epath.Path,
    name: str,
) -> Optional[str]:
  """Returns the default config of the given dataset, None if not found."""
  # Search for the DatasetBuilder generation code
  try:
    # Warning: The registered dataset may not match the files (e.g. if
    # the imported datasets has the same name as the generated files while
    # being 2 differents datasets)
    cls = registered.imported_builder_cls(name)
    cls = typing.cast(Type[dataset_builder.DatasetBuilder], cls)
  except registered.DatasetNotFoundError:
    pass
  except PermissionError as e:
    error_msg = f'Permission error when accessing: {builder_dir}: {e}'
    error_utils.add_context(error_msg)
  else:
    # If code found, return the default config
    if cls.BUILDER_CONFIGS:
      return cls.default_builder_config.name

  # Otherwise, try to load default config from common metadata
  return dataset_builder.load_default_config_name(epath.Path(builder_dir))


def _get_version_str(
    builder_dir: epath.Path,
    *,
    config_name: Optional[str] = None,
    requested_version: Optional[str] = None,
) -> Optional[str]:
  """Returns the version name found in the directory.

  Args:
    builder_dir: Directory containing the versions (`builder_dir/1.0.0/`,...)
    config_name: Optional name of the config that should be used. Will be
      ignored if it is an empty string.
    requested_version: Optional version to search (e.g. `1.0.0`, `2.*.*`,...)

  Returns:
    version_str: The version directory name found in `builder_dir`.
  """
  if config_name:
    builder_dir = builder_dir / config_name
  all_versions = version_lib.list_all_versions(os.fspath(builder_dir))
  # Version not given, using the latest one.
  if not requested_version and all_versions:
    return str(all_versions[-1])
  # Version given, return the highest version matching `requested_version`.
  for v in reversed(all_versions):
    if v.match(requested_version):
      return str(v)
  # Directory doesn't have version, or requested_version doesn't match
  if requested_version:
    error_msg = (
        f'No version matching the requested {requested_version} was '
        f'found in the builder directory: {builder_dir}.'
    )
  else:
    error_msg = (
        f"The builder directory {builder_dir} doesn't contain any versions."
    )
  error_utils.add_context(error_msg)
  return None



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""criteo dataset."""

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.recommendation.criteo import criteo


class CriteoTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for criteo dataset."""

  DATASET_CLASS = criteo.Criteo
  SPLITS = {'train': 5}  # Number of fake train example

  # If you are calling `download/download_and_extract` with a dict, like:
  #   dl_manager.download({'some_key': 'http://a.org/out.txt', ...})
  # then the tests needs to provide the fake output paths relative to the
  # fake data directory
  DL_EXTRACT_RESULT = 'fake_data.txt'


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Builder whose configs are transformations applied to other datasets.

Note that this is an experimental new feature, so the API may change.
"""

from __future__ import annotations

import functools
from typing import Callable, Dict, Iterator, List, Optional, Union

from tensorflow_datasets.core import beam_utils
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import dataset_utils
from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core import load as load_lib
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import registered
from tensorflow_datasets.core import split_builder as split_builder_lib
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core import transform as transform_lib
from tensorflow_datasets.core.download import download_manager
from tensorflow_datasets.core.utils import read_config as read_config_lib
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

Key = transform_lib.Key
Example = transform_lib.Example
KeyExample = transform_lib.KeyExample


class ViewConfig(dataset_builder.BuilderConfig):
  """Builder config for a view transforming another dataset."""

  def __init__(
      self,
      *,
      name: str,
      input_dataset: Union[None, str, naming.DatasetReference] = None,
      ex_transformations: Optional[
          List[transform_lib.ExampleTransformFn]
      ] = None,
      ds_transformations: Optional[
          List[Callable[[tf.data.Dataset], tf.data.Dataset]]
      ] = None,
      **kwargs,
  ):
    """Initialize a ViewConfig.

    Arguments:
      name: the name of this config.
      input_dataset: the dataset on which this view is being applied. If `None`,
        the input dataset needs to be specified in the `ViewBuilder`.
      ex_transformations: transformations that need to be applied to individual
        examples / rows. The transformations are normal Python functions. See
        the documentation for more information. Note that you cannot specify
        both example and dataset transformations.
      ds_transformations: `tf.data` transformations that are applied on the
        whole `tf.data.Dataset`. Note that you cannot specify both example and
        dataset transformations.
      **kwargs: additional arguments passed to the `BuilderConfig` such as
        `version`, `release_notes`, `supported_versions`, and `description`.
    """
    if isinstance(input_dataset, str):
      input_dataset = naming.DatasetReference.from_tfds_name(input_dataset)
    self.input_dataset = input_dataset
    if ex_transformations and ds_transformations:
      raise ValueError(
          "It is not supported to use both TF data and example transformations!"
      )
    self.ex_transformations = ex_transformations
    self.ds_transformations = ds_transformations
    super().__init__(name=name, **kwargs)

  def input_dataset_builder(
      self,
      data_dir: Optional[str] = None,
  ) -> dataset_builder.DatasetBuilder:
    assert self.input_dataset, "Input dataset must be specified."
    tfds_name = self.input_dataset.tfds_name(include_version=True)
    data_dir = self.input_dataset.data_dir or data_dir
    return load_lib.builder(name=tfds_name, data_dir=data_dir)


def _transform_per_example(
    builder: dataset_builder.DatasetBuilder,
    split_info: splits_lib.SplitInfo,
    transformations: List[transform_lib.ExampleTransformFn],
) -> split_builder_lib.SplitGenerator:
  """Yields transformed examples of the given builder and split.

  Arguments:
    builder: the builder whose data should be transformed.
    split_info: the split that needs to be transformed.
    transformations: the example transformations that need to be applied.

  Yields:
    transformed examples of the given builder and split.
  """
  row_read_config = read_config_lib.ReadConfig(add_tfds_id=True)
  for file_instruction in split_info.file_instructions:
    file_info = naming.FilenameInfo.from_str(file_instruction.filename)
    split_shard = f"{file_info.split}[{file_info.shard_index}shard]"
    ds = builder.as_dataset(split=split_shard, read_config=row_read_config)
    for example in dataset_utils.as_numpy(ds):
      key = example.pop("tfds_id")
      yield from transform_lib.apply_transformations(
          key=key, example=example, transformations=transformations
      )


def _transform_example(
    example: Example, transformations: List[transform_lib.ExampleTransformFn]
) -> Iterator[KeyExample]:
  key = example.pop("tfds_id")
  yield from transform_lib.apply_transformations(
      key=key, example=example, transformations=transformations
  )


def _transform_per_example_beam(
    builder: dataset_builder.DatasetBuilder,
    split_info: splits_lib.SplitInfo,
    transformations: List[transform_lib.ExampleTransformFn],
    workers_per_shard: int = 1,
) -> split_builder_lib.SplitGenerator:
  beam = lazy_imports_lib.lazy_imports.apache_beam
  split = split_info.name
  read_config = read_config_lib.ReadConfig(add_tfds_id=True)
  return (
      f"read_tfds_dataset@{split}"
      >> beam_utils.ReadFromTFDS(
          builder=builder,
          split=split,
          read_config=read_config,
          workers_per_shard=workers_per_shard,
      )
      | f"convert_to_numpy@{split}" >> beam.Map(dataset_utils.as_numpy)
      | f"transform_examples@{split}"
      >> beam.ParDo(_transform_example, transformations=transformations)
  )


def _transform_dataset(
    builder: dataset_builder.DatasetBuilder,
    split_info: splits_lib.SplitInfo,
    transformations: List[Callable[[tf.data.Dataset], tf.data.Dataset]],
) -> split_builder_lib.SplitGenerator:
  """Yields transformed examples of the given builder and split.

  Arguments:
    builder: the builder whose data should be transformed.
    split_info: the split that needs to be transformed.
    transformations: the dataset transformations that need to be applied.

  Yields:
    transformed examples of the given builder and split.
  """
  row_read_config = read_config_lib.ReadConfig(add_tfds_id=False)
  ds = builder.as_dataset(split=split_info.name, read_config=row_read_config)
  for transformation in transformations:
    ds = transformation(ds)
  for i, example in enumerate(dataset_utils.as_numpy(ds)):
    key = i
    yield key, example


class ViewBuilder(
    dataset_builder.GeneratorBasedBuilder, skip_registration=True
):
  """[Experimental] Base builder for views.

  Note that this is an experimental new feature, so the API may change.

  `ViewBuilder` can be used to define a new dataset as the transformation of an
  existing dataset.
  """

  # The dataset that this view transforms. The `input_dataset` in `ViewConfig`
  # takes precedence if specified.
  INPUT_DATASET: Union[None, str, naming.DatasetReference] = None

  # The transformations that need to be applied. Transformations in `ViewConfig`
  # take precedence if specified. Note that one should not specify both
  # `EX_TRANSFORMATIONS` and `DS_TRANSFORMATIONS`.
  # Transformations that are applied per example / row.
  EX_TRANSFORMATIONS: List[transform_lib.ExampleTransformFn] = []
  # Transformations that are applied on the full `tf.data.Dataset`.
  DS_TRANSFORMATIONS: List[Callable[[tf.data.Dataset], tf.data.Dataset]] = []

  # Whether to use Beam to generate this view.
  USE_BEAM = False

  # The number of workers that should read a single shard. If set to higher
  # than 1, multiple Beam workers will read a single shard. For tfrecord, that
  # means that some redundant reads are done.
  BEAM_WORKERS_PER_SHARD = 1

  @functools.cached_property
  def view_config(self) -> Optional[ViewConfig]:
    if isinstance(self.builder_config, ViewConfig):
      return self.builder_config
    return None

  def _src_dataset(self) -> naming.DatasetReference:
    input_dataset = self.view_config and self.view_config.input_dataset
    if input_dataset:
      return input_dataset
    elif self.INPUT_DATASET is not None:
      if isinstance(self.INPUT_DATASET, str):
        return naming.DatasetReference.from_tfds_name(self.INPUT_DATASET)
      return self.INPUT_DATASET
    raise ValueError("Input dataset was not specified!")

  def _src_dataset_builder(self) -> dataset_builder.DatasetBuilder:
    tfds_name = self._src_dataset().tfds_name(include_version=True)
    # First try to load the dataset from the data_dir where this dataset is
    # downloaded to. If that fails, load it from the default data dir.
    try:
      data_dir = self._src_dataset().data_dir or self._data_dir_root
      builder = load_lib.builder(name=tfds_name, data_dir=data_dir)
      if builder.is_prepared():
        return builder
    except registered.DatasetNotFoundError:
      pass
    return load_lib.builder(name=tfds_name)

  def _example_transformations(self) -> List[transform_lib.ExampleTransformFn]:
    ex_transformations = (
        self.view_config and self.view_config.ex_transformations
    )
    if ex_transformations:
      return ex_transformations
    return self.EX_TRANSFORMATIONS

  def _data_transformations(
      self,
  ) -> List[Callable[[tf.data.Dataset], tf.data.Dataset]]:
    if self.view_config and self.view_config.ds_transformations:
      return self.view_config.ds_transformations
    return self.DS_TRANSFORMATIONS

  def _split_generators(
      self, dl_manager: download_manager.DownloadManager
  ) -> Dict[splits_lib.Split, split_builder_lib.SplitGenerator]:
    del dl_manager
    builder = self._src_dataset_builder()
    split_generators = {}
    for split, split_info in builder.info.splits.items():
      split_generators[split] = self._generate_examples(
          builder=builder,
          split_info=split_info,
          ex_transformations=self._example_transformations(),
          ds_transformations=self._data_transformations(),
      )
    return split_generators

  def _generate_examples(
      self,
      builder: dataset_builder.DatasetBuilder,
      split_info: splits_lib.SplitInfo,
      ex_transformations: List[transform_lib.ExampleTransformFn],
      ds_transformations: List[Callable[[tf.data.Dataset], tf.data.Dataset]],
  ) -> split_builder_lib.SplitGenerator:
    if ex_transformations and ds_transformations:
      raise ValueError(
          "It is not supported to have toth example and dataset "
          "transformations!"
      )
    if ex_transformations:
      if self.USE_BEAM:
        return _transform_per_example_beam(
            builder=builder,
            split_info=split_info,
            transformations=ex_transformations,
            workers_per_shard=self.BEAM_WORKERS_PER_SHARD,
        )
      else:
        return _transform_per_example(
            builder=builder,
            split_info=split_info,
            transformations=ex_transformations,
        )
    elif ds_transformations:
      if self.USE_BEAM:
        # TODO(weide): add support for using Beam with tf.data transformations.
        raise ValueError(
            "Using Apache Beam in combination with tf.data "
            "transformations is not yet supported!"
        )
      return _transform_dataset(
          builder=builder,
          split_info=split_info,
          transformations=ds_transformations,
      )
    raise ValueError("No transformations were specified!")



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""rlu_locomotion dataset."""

from __future__ import annotations

from typing import Any, Dict, Optional

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.rl_unplugged import rlu_common

_LOCOMOTION_DESCRIPTION = """
These tasks are made up of the corridor locomotion tasks involving the CMU
Humanoid, for which prior efforts have either used motion capture data
[Merel et al., 2019a](https://arxiv.org/abs/1811.09656),
[Merel et al., 2019b](https://arxiv.org/abs/1811.11711) or training from scratch
[Song et al., 2020](https://arxiv.org/abs/1909.12238). In addition, the DM
Locomotion repository contains a set of tasks adapted to be suited to a virtual
rodent [Merel et al., 2020](https://arxiv.org/abs/1911.09451). We emphasize that
the DM Locomotion tasks feature the combination of challenging high-DoF
continuous control along with perception from rich egocentric observations.
For details on how the dataset was generated, please refer to the paper.

We recommend you to try offline RL methods on DeepMind Locomotion dataset, if
you are interested in very challenging offline RL dataset with continuous action
space.
"""

_CITATION = """
@inproceedings{gulcehre2020rl,
 title = {RL Unplugged: A Suite of Benchmarks for Offline Reinforcement Learning},
 author = {Gulcehre, Caglar and Wang, Ziyu and Novikov, Alexander and Paine, Thomas and G\'{o}mez, Sergio and Zolna, Konrad and Agarwal, Rishabh and Merel, Josh S and Mankowitz, Daniel J and Paduraru, Cosmin and Dulac-Arnold, Gabriel and Li, Jerry and Norouzi, Mohammad and Hoffman, Matthew and Heess, Nicolas and de Freitas, Nando},
 booktitle = {Advances in Neural Information Processing Systems},
 pages = {7248--7259},
 volume = {33},
 year = {2020}
}
"""

_TASK_NAMES = [
    'humanoid_corridor',
    'humanoid_gaps',
    'humanoid_walls',
    'rodent_bowl_escape',
    'rodent_gaps',
    'rodent_mazes',
    'rodent_two_touch',
]


def _sequence(
    shape_size: Optional[int] = None, dtype: Optional[tf.DType] = None
) -> tf.io.FixedLenSequenceFeature:
  if dtype is None:
    dtype = tf.float32
  if shape_size:
    shape = [shape_size]
  else:
    shape = []
  return tf.io.FixedLenSequenceFeature(shape, dtype, allow_missing=True)


def _feature_description(task_name: str) -> Dict[str, Any]:
  """Returns the shape of a tf.Example for the given task."""
  if 'humanoid' in task_name:
    return {
        'observation/walker/joints_vel': _sequence(56),
        'observation/walker/sensors_velocimeter': _sequence(3),
        'observation/walker/sensors_gyro': _sequence(3),
        'observation/walker/joints_pos': _sequence(56),
        'observation/walker/world_zaxis': _sequence(3),
        'observation/walker/body_height': _sequence(1),
        'observation/walker/sensors_accelerometer': _sequence(3),
        'observation/walker/end_effectors_pos': _sequence(12),
        'observation/walker/egocentric_camera': _sequence(
            shape_size=None, dtype=tf.string
        ),
        'action': _sequence(56),
        'discount': _sequence(),
        'reward': _sequence(),
        'step_type': _sequence(),
        'episode_id': tf.io.FixedLenFeature([], tf.int64),
        'timestamp': tf.io.FixedLenFeature([], tf.int64),
    }
  else:
    return {
        'observation/walker/joints_pos': _sequence(30),
        'observation/walker/joints_vel': _sequence(30),
        'observation/walker/tendons_pos': _sequence(8),
        'observation/walker/tendons_vel': _sequence(8),
        'observation/walker/appendages_pos': _sequence(15),
        'observation/walker/world_zaxis': _sequence(3),
        'observation/walker/sensors_accelerometer': _sequence(3),
        'observation/walker/sensors_velocimeter': _sequence(3),
        'observation/walker/sensors_gyro': _sequence(3),
        'observation/walker/sensors_touch': _sequence(4),
        'observation/walker/egocentric_camera': _sequence(
            shape_size=None, dtype=tf.string
        ),
        'action': _sequence(38),
        'discount': _sequence(),
        'reward': _sequence(),
        'step_type': _sequence(),
        'episode_id': tf.io.FixedLenFeature([], tf.int64),
        'timestamp': tf.io.FixedLenFeature([], tf.int64),
    }


class RluLocomotion(rlu_common.RLUBuilder):
  """DatasetBuilder for rlu_locomotion dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  # pytype: disable=wrong-keyword-args
  BUILDER_CONFIGS = [tfds.core.BuilderConfig(name=name) for name in _TASK_NAMES]
  # pytype: enable=wrong-keyword-args

  _SHARDS = 100
  _INPUT_FILE_PREFIX = 'gs://rl_unplugged/dm_locomotion_episodes/'

  def get_features_dict(self):
    if 'humanoid' in self.builder_config.name:
      walker_features = {
          'joints_vel': rlu_common.float_tensor_feature(56),
          'sensors_velocimeter': rlu_common.float_tensor_feature(3),
          'sensors_gyro': rlu_common.float_tensor_feature(3),
          'joints_pos': rlu_common.float_tensor_feature(56),
          'world_zaxis': rlu_common.float_tensor_feature(3),
          'body_height': rlu_common.float_tensor_feature(1),
          'sensors_accelerometer': rlu_common.float_tensor_feature(3),
          'end_effectors_pos': rlu_common.float_tensor_feature(12),
          'egocentric_camera': tfds.features.Image(
              shape=(64, 64, 3), dtype=np.uint8, encoding_format='png'
          ),
      }
      action_features = tfds.features.Tensor(shape=(56,), dtype=np.float32)
    else:  # 'rodent' datasets
      walker_features = {
          'joints_pos': rlu_common.float_tensor_feature(30),
          'joints_vel': rlu_common.float_tensor_feature(30),
          'tendons_pos': rlu_common.float_tensor_feature(8),
          'tendons_vel': rlu_common.float_tensor_feature(8),
          'appendages_pos': rlu_common.float_tensor_feature(15),
          'world_zaxis': rlu_common.float_tensor_feature(3),
          'sensors_accelerometer': rlu_common.float_tensor_feature(3),
          'sensors_velocimeter': rlu_common.float_tensor_feature(3),
          'sensors_gyro': rlu_common.float_tensor_feature(3),
          'sensors_touch': rlu_common.float_tensor_feature(4),
          'egocentric_camera': tfds.features.Image(
              shape=(64, 64, 3), dtype=np.uint8, encoding_format='png'
          ),
      }
      action_features = tfds.features.Tensor(shape=(38,), dtype=np.float32)

    return tfds.features.FeaturesDict({
        'steps': tfds.features.Dataset({
            'observation': {
                'walker': walker_features,
            },
            'action': action_features,
            'reward': np.float32,
            'is_terminal': np.bool_,
            'is_first': np.bool_,
            'is_last': np.bool_,
            'discount': np.float32,
        }),
        'episode_id': np.int64,
        'timestamp': np.int64,
    })

  def get_description(self):
    return _LOCOMOTION_DESCRIPTION

  def get_citation(self):
    return _CITATION

  def get_file_prefix(self):
    task = self.builder_config.name
    return f'{self._INPUT_FILE_PREFIX}/{task}/train'

  def num_shards(self):
    return self._SHARDS

  def tf_example_to_step_ds(
      self, tf_example: tf.train.Example
  ) -> Dict[str, Any]:
    """Create an episode from a TF example."""
    feature_description = _feature_description(self.builder_config.name)

    data = tf.io.parse_single_example(tf_example, feature_description)
    episode_length = tf.size(data['discount'])
    is_first = tf.concat(
        [[True], [False] * tf.ones(episode_length - 1)], axis=0
    )
    is_last = tf.concat([[False] * tf.ones(episode_length - 1), [True]], axis=0)
    is_terminal = [False] * tf.ones(episode_length, tf.int64)

    # The data is in RSA alignment, we realign it to SAR to comply with the
    # RLDS standard.
    discount = data['discount'][1:]
    if discount[-1] == 0.0:
      is_terminal = tf.concat(
          [[False] * tf.ones(episode_length - 1, tf.int64), [True]], axis=0
      )
      # If the episode ends in a terminal state, in the last step only the
      # observation has valid information (the terminal state).
      discount = tf.concat([discount, [0.0]], axis=0)
    else:
      discount = tf.concat([discount, [1.0]], axis=0)

    reward = tf.concat([data['reward'][1:], [0.0]], axis=0)

    obs_prefix = 'observation/walker/'
    obs = {}
    for k in feature_description:
      if k.startswith(obs_prefix):
        new_k = k[len(obs_prefix) :]
        if 'egocentric_camera' in k:
          obs[new_k] = tf.reshape(
              tf.io.decode_raw(data[k], out_type=tf.uint8),
              (-1,)
              + (
                  64,
                  64,
                  3,
              ),
          )
        else:
          obs[new_k] = data[k]
    episode = {
        'steps': {
            'observation': {'walker': obs},
            'action': data['action'],
            'reward': reward,
            'discount': discount,
            'is_first': is_first,
            'is_last': is_last,
            'is_terminal': is_terminal,
        },
        'episode_id': data['episode_id'],
        'timestamp': data['timestamp'],
    }
    return episode



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""wikiann dataset."""

import os

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """\
WikiANN (sometimes called PAN-X) is a multilingual named entity recognition \
dataset consisting of Wikipedia articles annotated with LOC (location), PER \
(person), and ORG (organisation) tags in the IOB2 format. This version \
corresponds to the balanced train, dev, and test splits of Rahimi et al. \
(2019), which supports 176 of the 282 languages from the original WikiANN \
corpus.
"""

_CITATION = """
@inproceedings{rahimi-etal-2019-massively,
    title = "Massively Multilingual Transfer for {NER}",
    author = "Rahimi, Afshin  and
      Li, Yuan  and
      Cohn, Trevor",
    booktitle = "Proceedings of the 57th Annual Meeting of the Association \
    for Computational Linguistics",
    month = jul,
    year = "2019",
    address = "Florence, Italy",
    publisher = "Association for Computational Linguistics",
    url = "https://www.aclweb.org/anthology/P19-1015",
    pages = "151--164",
}
"""

URL = "https://www.dropbox.com/s/12h3qqog6q4bjve/panx_dataset.tar?dl=1"

LANGS = [
    "ace",
    "af",
    "als",
    "am",
    "ang",
    "an",
    "arc",
    "ar",
    "arz",
    "as",
    "ast",
    "ay",
    "az",
    "bar",
    "ba",
    "bat-smg",
    "be",
    "be-x-old",
    "bg",
    "bh",
    "bn",
    "bo",
    "br",
    "bs",
    "ca",
    "cbk-zam",
    "cdo",
    "ceb",
    "ce",
    "ckb",
    "co",
    "crh",
    "csb",
    "cs",
    "cv",
    "cy",
    "da",
    "de",
    "diq",
    "dv",
    "el",
    "eml",
    "en",
    "eo",
    "es",
    "et",
    "eu",
    "ext",
    "fa",
    "fi",
    "fiu-vro",
    "fo",
    "frr",
    "fr",
    "fur",
    "fy",
    "gan",
    "ga",
    "gd",
    "gl",
    "gn",
    "gu",
    "hak",
    "he",
    "hi",
    "hr",
    "hsb",
    "hu",
    "hy",
    "ia",
    "id",
    "ig",
    "ilo",
    "io",
    "is",
    "it",
    "ja",
    "jbo",
    "jv",
    "ka",
    "kk",
    "km",
    "kn",
    "ko",
    "ksh",
    "ku",
    "ky",
    "la",
    "lb",
    "lij",
    "li",
    "lmo",
    "ln",
    "lt",
    "lv",
    "map-bms",
    "mg",
    "mhr",
    "min",
    "mi",
    "mk",
    "ml",
    "mn",
    "mr",
    "ms",
    "mt",
    "mwl",
    "my",
    "mzn",
    "nap",
    "nds",
    "ne",
    "nl",
    "nn",
    "no",
    "nov",
    "oc",
    "or",
    "os",
    "pa",
    "pdc",
    "pl",
    "pms",
    "pnb",
    "ps",
    "pt",
    "qu",
    "rm",
    "ro",
    "ru",
    "rw",
    "sah",
    "sa",
    "scn",
    "sco",
    "sd",
    "sh",
    "simple",
    "si",
    "sk",
    "sl",
    "so",
    "sq",
    "sr",
    "su",
    "sv",
    "sw",
    "szl",
    "ta",
    "te",
    "tg",
    "th",
    "tk",
    "tl",
    "tr",
    "tt",
    "ug",
    "uk",
    "ur",
    "uz",
    "vec",
    "vep",
    "vi",
    "vls",
    "vo",
    "war",
    "wa",
    "wuu",
    "xmf",
    "yi",
    "yo",
    "zea",
    "zh-classical",
    "zh-min-nan",
    "zh",
    "zh-yue",
]


class WikiannConfig(tfds.core.BuilderConfig):
  """ConfigurationClass for wikiann dataset."""

  def __init__(self, *, language, **kwargs):
    if language not in LANGS:
      raise ValueError("language must be one of {}".format(list(LANGS)))

    super(WikiannConfig, self).__init__(**kwargs)
    self.language = language


def tags_to_spans(tags):
  """Convert tags to spans."""
  spans = set()
  span_start = 0
  span_end = 0
  active_conll_tag = None
  for index, string_tag in enumerate(tags):
    # Actual BIO tag.
    bio_tag = string_tag[0]
    assert bio_tag in ["B", "I", "O"], "Invalid Tag"
    conll_tag = string_tag[2:]
    if bio_tag == "O":
      # The span has ended.
      if active_conll_tag:
        spans.add((active_conll_tag, (span_start, span_end)))
      active_conll_tag = None
      # We don't care about tags we are
      # told to ignore, so we do nothing.
      continue
    elif bio_tag == "B":
      # We are entering a new span; reset indices and active tag to new span.
      if active_conll_tag:
        spans.add((active_conll_tag, (span_start, span_end)))
      active_conll_tag = conll_tag
      span_start = index
      span_end = index
    elif bio_tag == "I" and conll_tag == active_conll_tag:
      # We're inside a span.
      span_end += 1
    else:
      # This is the case the bio label is an "I", but either:
      # 1) the span hasn't started - i.e. an ill formed span.
      # 2) We have IOB1 tagging scheme.
      # We'll process the previous span if it exists, but also include this
      # span. This is important, because otherwise, a model may get a perfect
      # F1 score whilst still including false positive ill-formed spans.
      if active_conll_tag:
        spans.add((active_conll_tag, (span_start, span_end)))
      active_conll_tag = conll_tag
      span_start = index
      span_end = index
  # Last token might have been a part of a valid span.
  if active_conll_tag:
    spans.add((active_conll_tag, (span_start, span_end)))
  # Return sorted list of spans
  return sorted(list(spans), key=lambda x: x[1][0])


def get_spans(tokens, tags):
  """Convert tags to textspans."""
  spans = tags_to_spans(tags)
  text_spans = [
      x[0] + ": " + " ".join([tokens[i] for i in range(x[1][0], x[1][1] + 1)])
      for x in spans
  ]
  if not text_spans:
    text_spans = ["None"]
  return text_spans


class Wikiann(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for wikiann dataset."""

  BUILDER_CONFIGS = [
      WikiannConfig(  # pylint: disable=g-complex-comprehension
          name=language,
          description="Wikiann {} train/dev/test splits".format(language),
          version="1.0.0",
          language=language,
      )
      for language in LANGS
  ]

  VERSION = tfds.core.Version("1.0.0")
  RELEASE_NOTES = {
      "1.0.0": "Initial release.",
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    features = tfds.features.FeaturesDict({
        "tokens": tfds.features.Sequence(tfds.features.Text()),
        "tags": tfds.features.Sequence(
            tfds.features.ClassLabel(
                names=[
                    "O",
                    "B-PER",
                    "I-PER",
                    "B-ORG",
                    "I-ORG",
                    "B-LOC",
                    "I-LOC",
                ]
            )
        ),
        "langs": tfds.features.Sequence(tfds.features.Text()),
        "spans": tfds.features.Sequence(tfds.features.Text()),
    })
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=features,
        supervised_keys=None,
        homepage="https://github.com/afshinrahimi/mmner",
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    path = dl_manager.download_and_extract(URL)
    subpath = dl_manager.extract(
        os.path.join(path, self.builder_config.language + ".tar.gz")
    )

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={"filepath": os.path.join(subpath, "dev")},
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={"filepath": os.path.join(subpath, "test")},
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={"filepath": os.path.join(subpath, "train")},
        ),
    ]

  def _generate_examples(self, filepath):
    """Reads line by line format of the NER dataset and generates examples.

    Input Format:

    en:rick  B-PER
    en:and  O
    en:morty  B-PER
    en:are  O
    en:cool  O
    en:.  O

    Output Format:

    {
    'tokens': ["rick", "and", "morty", "are", "cool", "."],
    'tags': ["B-PER", "O" , "B-PER", "O", "O", "O"],
    'langs': ["en", "en", "en", "en", "en", "en"]
    'spans': ["PER: rick", "PER: morty"]
    }

    Args:
      filepath: Path to file with line by line NER format.

    Yields:
      Examples with the format listed above.
    """

    key = 1
    with tf.io.gfile.GFile(filepath, "r") as f:
      tokens = []
      tags = []
      langs = []
      for line in f:
        line = line.rstrip()
        # pylint: disable=g-explicit-bool-comparison
        if line.startswith("-DOCSTART-") or line == "":
          if tokens:
            spans = get_spans(tokens, tags)
            yield key, {
                "tokens": tokens,
                "tags": tags,
                "langs": langs,
                "spans": spans,
            }
            key += 1
            tokens = []
            tags = []
            langs = []
        else:
          # wikiann data is tab separated
          fields = line.split("\t")
          # strip out language prefix
          langs.append(fields[0].split(":")[0])
          tokens.append(":".join(fields[0].split(":")[1:]))
          if len(fields) > 1:
            tags.append(fields[-1])
          else:
            # examples have no label in test set
            tags.append("O")
      if tokens:
        spans = get_spans(tokens, tags)
        yield key, {
            "tokens": tokens,
            "tags": tags,
            "langs": langs,
            "spans": spans,
        }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Cache utils.

The cache (default to `~/.cache/tensorflow_datasets/`) is used for:

* The community dataset index list (cached in
  `<cache_dir>/community-datasets-list.jsonl` from
  `gs://tfds-data/community-datasets-list.jsonl`)
* The installed dataset packages (downloaded from github and installed in
  `<cache_dir>/modules/tfds_community/`).

"""

import os
import sys

from etils import epath
from tensorflow_datasets.core import utils


def _default_cache_dir() -> epath.Path:
  """Returns the default cache directory."""
  if 'TFDS_CACHE_DIR' in os.environ:
    path = os.environ['TFDS_CACHE_DIR']
  elif 'XDG_CACHE_HOME' in os.environ:
    path = os.path.join(os.environ['XDG_CACHE_HOME'], 'tensorflow_datasets')
  else:
    path = os.path.join('~', '.cache', 'tensorflow_datasets')
  return epath.Path(path).expanduser()


@utils.memoize()
def cache_path() -> epath.Path:
  """Returns the path to the TFDS cache."""
  path = _default_cache_dir()
  path.mkdir(parents=True, exist_ok=True)
  return path


@utils.memoize()
def module_path() -> epath.Path:
  """Returns the path to the cached TFDS dynamically installed modules.

  Calling this function will update `sys.path` so modules installed in this
  directory can be imported.

  Returns:
    module_path: The path to the dynamically installed modules.
  """
  path = cache_path() / 'modules/'
  path.mkdir(parents=True, exist_ok=True)
  # Add the `~/.cache/tensorflow_datasets/modules/` to `sys.path`
  sys.path.append(os.fspath(path))
  return path



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CIFAR100-N dataset."""

from __future__ import annotations

import collections
import os

from etils import epath
import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

# Shared constants
_CIFAR_IMAGE_SIZE = 32
_CIFAR_IMAGE_SHAPE = (_CIFAR_IMAGE_SIZE, _CIFAR_IMAGE_SIZE, 3)

_DESCRIPTION = """
A re-labeled version of CIFAR-100 with real human annotation errors. For every 
pair (image, label) in the original CIFAR-100 train set, it provides an 
additional label given by a real human annotator. 
"""

_CITATION = """
@inproceedings{wei2022learning,
  title={Learning with Noisy Labels Revisited: A Study Using Real-World Human 
  Annotations},
  author={Jiaheng Wei and Zhaowei Zhu and Hao Cheng and Tongliang Liu and Gang 
  Niu and Yang Liu},
  booktitle={International Conference on Learning Representations},
  year={2022},
  url={https://openreview.net/forum?id=TBWA6PLJZQm}
}
"""


class Cifar100N(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for CIFAR-10N dataset."""

  MANUAL_DOWNLOAD_INSTRUCTIONS = """
  Download 'side_info_cifar100N.csv', 'CIFAR-100_human_ordered.npy' and
  'image_order_c100.npy' from https://github.com/UCSC-REAL/cifar-10-100n.
  
  Then convert 'CIFAR-100_human_ordered.npy' into a CSV file 
  'CIFAR-100_human_annotations.csv'. This can be done with the following code:
  
  ```
  import numpy as np
  from tensorflow_datasets.core.utils.lazy_imports_utils import pandas as pd
  from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
  
  human_labels_np_path = '<local_path>/CIFAR-100_human_ordered.npy'
  human_labels_csv_path = '<local_path>/CIFAR-100_human_annotations.csv'

  with tf.io.gfile.GFile(human_labels_np_path, "rb") as f:
    human_annotations = np.load(f, allow_pickle=True)

  df = pd.DataFrame(human_annotations[()])

  with tf.io.gfile.GFile(human_labels_csv_path, "w") as f:
    df.to_csv(f, index=False)
  ```
  """

  VERSION = tfds.core.Version('1.0.1')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
      '1.0.1': 'Fixed correspondence between annotations and images.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            'id': tfds.features.Text(),
            'image': tfds.features.Image(shape=_CIFAR_IMAGE_SHAPE),
            'label': tfds.features.ClassLabel(num_classes=100),
            'coarse_label': tfds.features.ClassLabel(num_classes=20),
            'noise_label': tfds.features.ClassLabel(num_classes=100),
            'worker_id': np.int64,
            'worker_time': np.float32,
        }),
        supervised_keys=None,
        homepage='https://www.cs.toronto.edu/~kriz/cifar.html',
        citation=_CITATION,
    )

  @property
  def _cifar_info(self):
    return CifarInfo(
        name=self.name,
        url='https://www.cs.toronto.edu/~kriz/cifar-100-binary.tar.gz',
        train_files=['train.bin'],
        test_files=['test.bin'],
        prefix='cifar-100-binary/',
        label_files=['coarse_label_names.txt', 'fine_label_names.txt'],
        label_keys=['coarse_label', 'label'],
        human_label_path='CIFAR-100_human_annotations.csv',
        side_info_path='side_info_cifar100N.csv',
        annotations_order_path='image_order_c100.npy',
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    cifar_path = dl_manager.download_and_extract(self._cifar_info.url)
    cifar_info = self._cifar_info

    cifar_path = os.path.join(cifar_path, cifar_info.prefix)

    dl_paths = {
        'human_labels': dl_manager.manual_dir / cifar_info.human_label_path,
        'side_info': dl_manager.manual_dir / cifar_info.side_info_path,
        'annotations_order': (
            dl_manager.manual_dir / cifar_info.annotations_order_path
        ),
    }

    # Load the label names
    for label_key, label_file in zip(
        cifar_info.label_keys, cifar_info.label_files
    ):
      labels_path = os.path.join(cifar_path, label_file)
      with epath.Path(labels_path).open() as label_f:
        label_names = [name for name in label_f.read().split('\n') if name]
      self.info.features[label_key].names = label_names

    # Define the splits
    def gen_filenames(filenames):
      for f in filenames:
        yield os.path.join(cifar_path, f)

    return {
        'train': self._generate_examples(
            'train_', gen_filenames(cifar_info.train_files), dl_paths
        ),
        'test': self._generate_examples(
            'test_', gen_filenames(cifar_info.test_files), dl_paths
        ),
    }

  def _generate_examples(self, split_prefix, filepaths, dl_paths):
    """Generate CIFAR-100N examples as dicts.

    Uses self._cifar_info as configuration.

    Args:
      split_prefix (str): Prefix that identifies the split (e.g. "tr" or "te").
      filepaths (list[str]): The files to use to generate the data.
      dl_paths (dict[str, str]): Paths to human_labels and side_info

    Yields:
      The cifar examples, as defined in the dataset info features.
    """
    label_keys = self._cifar_info.label_keys
    index = 0  # Using index as key since data is always loaded in same order.

    if 'train' in split_prefix:
      human_labels = _load_human_labels(dl_paths['human_labels'])
      noise_label = human_labels['noise_label']

      side_info = _load_side_info(dl_paths['side_info'])
      worker_id = side_info['worker_id']
      worker_time = side_info['worker_time']

      annotations_order = _load_annotations_order(dl_paths['annotations_order'])

    for path in filepaths:
      for labels, np_image in _load_data(path, len(label_keys)):
        record = dict(zip(label_keys, labels))
        # Note: "id" is only provided for the user convenience. To shuffle the
        # dataset we use `index`, so that the sharding is compatible with
        # earlier versions.
        record['id'] = '{}{:05d}'.format(split_prefix, index)
        record['image'] = np_image

        if 'train' in split_prefix:
          # Note: The human labels are provided according to default shuffling
          # of 'cifar100'. We need to invert the shuffling to assign the right
          # human label to each example.
          annotations_index = np.where(annotations_order == index)[0][0]

          record['noise_label'] = noise_label[annotations_index]
          # Worker metadata is shared every 5 samples
          record['worker_id'] = worker_id[index // 5]
          record['worker_time'] = worker_time[index // 5]
        else:
          # There is no annotator metadata for test split
          record['noise_label'] = -1
          record['worker_id'] = -1
          record['worker_time'] = -1

        yield index, record
        index += 1


class CifarInfo(
    collections.namedtuple(
        '_CifarInfo',
        [
            'name',
            'url',
            'prefix',
            'train_files',
            'test_files',
            'label_files',
            'label_keys',
            'human_label_path',
            'side_info_path',
            'annotations_order_path',
        ],
    )
):
  """Contains the information necessary to generate a CIFAR dataset.

  Attributes:
    name (str): name of dataset.
    url (str): data URL.
    prefix (str): path prefix within the downloaded and extracted file to look
      for `train_files` and `test_files`.
    train_files (list<str>): name of training files within `prefix`.
    test_files (list<str>): name of test files within `prefix`.
    label_files (list<str>): names of the label files in the data.
    label_keys (list<str>): names of the label keys in the data.
    human_label_path (str): path to human annotations.
    side_info_path (str): path to metadata about annotations.
    annotations_order_path (str): path to annotation-image order correspondence.
  """


def _load_data(path, labels_number=1):
  """Yields (labels, np_image) tuples."""
  with tf.io.gfile.GFile(path, 'rb') as f:
    data = f.read()
  offset = 0
  max_offset = len(data) - 1
  while offset < max_offset:
    labels = np.frombuffer(
        data, dtype=np.uint8, count=labels_number, offset=offset
    ).reshape((labels_number,))
    # 1 byte per label, 1024 * 3 = 3072 bytes for the image.
    offset += labels_number
    img = (
        np.frombuffer(data, dtype=np.uint8, count=3072, offset=offset)
        .reshape((3, _CIFAR_IMAGE_SIZE, _CIFAR_IMAGE_SIZE))
        .transpose((1, 2, 0))
    )
    offset += 3072
    yield labels, img


def _load_annotations_order(path):
  """Loads index mapping between the annotation files and the CIFAR100 binaries."""
  with tf.io.gfile.GFile(path, 'rb') as f:
    annotations_order = np.load(f)
  return annotations_order


def _load_side_info(path):
  """Loads information from side_info_cifar100N.csv."""
  side_info_key_map = {
      1: 'worker_id',
      2: 'worker_time',
  }
  side_info_array = np.genfromtxt(
      tf.io.gfile.GFile(path), delimiter=',', skip_header=1
  )
  side_info = {}
  for key in side_info_key_map:
    side_info[side_info_key_map[key]] = side_info_array[:, key]
  return side_info


def _load_human_labels(path):
  """Loads information from side_info_cifar100N.csv."""
  with tf.io.gfile.GFile(path, 'r') as f:
    human_labels_array = np.genfromtxt(f, delimiter=',', skip_header=1)
  human_labels = {}
  human_labels['noise_label'] = human_labels_array[:, 1]
  return human_labels



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.features.feature."""

import pathlib
import sys
import textwrap
from unittest import mock

from absl import logging
from absl.testing import parameterized
import numpy as np
import tensorflow as tf
from tensorflow_datasets import testing
from tensorflow_datasets.core import features as features_lib


class AnInputConnector(features_lib.FeatureConnector):
  """Simple FeatureConnector implementing the based methods used for test."""

  def get_tensor_info(self):
    # With this connector, the way the data is on disk ({'a', 'b'}) do not match
    # the way it is exposed to the user (int64), so we overwrite
    # FeaturesDict.get_tensor_info
    return features_lib.TensorInfo(shape=(), dtype=np.int64)

  def get_serialized_info(self):
    return {
        'a': features_lib.TensorInfo(shape=(), dtype=np.int64),
        'b': features_lib.TensorInfo(shape=(), dtype=np.int64),
    }

  def encode_example(self, example_data):
    # Encode take the input data and wrap in in a dict
    return {'a': example_data + 1, 'b': example_data * 10}

  def decode_example(self, tfexample_dict):
    # Merge the two values
    return tfexample_dict['a'] + tfexample_dict['b']


class AnOutputConnector(features_lib.FeatureConnector):
  """Simple FeatureConnector implementing the based methods used for test."""

  def get_tensor_info(self):
    return features_lib.TensorInfo(shape=(), dtype=np.float32)

  def encode_example(self, example_data):
    return example_data * 10.0

  def decode_example(self, tfexample_data):
    return tfexample_data / 10.0


class FeatureDictTest(
    parameterized.TestCase, testing.FeatureExpectationsTestCase
):

  def test_tensor_info(self):
    self.assertEqual(
        features_lib.TensorInfo(shape=(None, 3), dtype=tf.string),
        features_lib.TensorInfo(shape=(None, 3), dtype=tf.string),
    )

    self.assertEqual(
        features_lib.TensorInfo(shape=(None, 3), dtype=np.object_),
        features_lib.TensorInfo(shape=(None, 3), dtype=np.object_),
    )

    self.assertNotEqual(
        features_lib.TensorInfo(shape=(None, 3), dtype=tf.string),
        features_lib.TensorInfo(shape=(None, 3), dtype=tf.int32),
    )

    self.assertNotEqual(
        features_lib.TensorInfo(shape=(2, 3), dtype=tf.string),
        features_lib.TensorInfo(shape=(5, 3), dtype=tf.string),
    )

    self.assertNotEqual(
        features_lib.TensorInfo(shape=(2, 3), dtype=np.object_),
        features_lib.TensorInfo(shape=(5, 3), dtype=np.object_),
    )

    t = features_lib.TensorInfo(shape=(None, 3), dtype=tf.string)
    self.assertEqual(t, features_lib.TensorInfo.copy_from(t))

  @parameterized.parameters(
      (
          np.int64,
          np.str_,
          mock.Mock(
              side_effect=Exception(
                  'no warning should be raised when using NumPy types'
              )
          ),
      ),
      (tf.int64, tf.string, mock.Mock()),
  )
  def test_tensor_spec(self, image_dtype, metadata_dtype, warning_mock):
    logging.warning = warning_mock
    feature = features_lib.FeaturesDict({
        'input': AnInputConnector(),
        'output': AnOutputConnector(),
        'img': {
            'size': {
                'height': features_lib.Tensor(shape=(2, 3), dtype=image_dtype),
                'width': features_lib.Tensor(
                    shape=[None, 3], dtype=image_dtype
                ),
            },
            'image': features_lib.Image(shape=(28, 28, 1)),
            'metadata/path': metadata_dtype,
        },
    })
    self.assertAllEqualNested(
        feature.get_tensor_spec(),
        {
            'input': tf.TensorSpec(shape=[], dtype=np.int64),
            'output': tf.TensorSpec(shape=[], dtype=np.float32),
            'img': {
                'size': {
                    'height': tf.TensorSpec(shape=[2, 3], dtype=np.int64),
                    'width': tf.TensorSpec(shape=[None, 3], dtype=np.int64),
                },
                'image': tf.TensorSpec(shape=[28, 28, 1], dtype=np.uint8),
                'metadata/path': tf.TensorSpec(shape=[], dtype=np.object_),
            },
        },
    )

  @parameterized.parameters(
      (np.int64, np.str_, np.object_), (tf.int64, tf.string, np.object_)
  )
  def test_fdict(self, image_dtype, metadata_dtype, output_metadata_dtype):
    self.assertFeature(
        feature=features_lib.FeaturesDict({
            'input': AnInputConnector(),
            'output': AnOutputConnector(),
            'img': {
                'size': {
                    'height': image_dtype,
                    'width': image_dtype,
                },
                'metadata/path': metadata_dtype,
            },
        }),
        serialized_info={
            'input': {
                'a': features_lib.TensorInfo(shape=(), dtype=np.int64),
                'b': features_lib.TensorInfo(shape=(), dtype=np.int64),
            },
            'output': features_lib.TensorInfo(shape=(), dtype=np.float32),
            'img': {
                'size': {
                    'height': features_lib.TensorInfo(shape=(), dtype=np.int64),
                    'width': features_lib.TensorInfo(shape=(), dtype=np.int64),
                },
                'metadata/path': features_lib.TensorInfo(
                    shape=(), dtype=output_metadata_dtype
                ),
            },
        },
        dtype={
            'input': np.int64,
            'output': np.float32,
            'img': {
                'size': {
                    'height': np.int64,
                    'width': np.int64,
                },
                'metadata/path': output_metadata_dtype,
            },
        },
        shape={
            'input': (),
            'output': (),
            'img': {
                'size': {
                    'height': (),
                    'width': (),
                },
                'metadata/path': (),
            },
        },
        tests=[
            # Np array
            testing.FeatureExpectationItem(
                value={
                    'input': 1,
                    'output': -1,
                    'img': {
                        'size': {
                            'height': 256,
                            'width': 128,
                        },
                        'metadata/path': 'path/to/xyz.jpg',
                    },
                },
                expected_serialized={
                    'input': {
                        'a': 2,  # 1 + 1
                        'b': 10,  # 1 * 10
                    },
                    'output': -10.0,  # -1 * 10.0
                    'img': {
                        'size': {
                            'height': 256,
                            'width': 128,
                        },
                        'metadata/path': 'path/to/xyz.jpg',
                    },
                },
                expected={
                    # a = 1 + 1, b = 1 * 10 => output = a + b = 2 + 10 = 12
                    'input': 12,  # 2 + 10
                    'output': -1.0,
                    'img': {
                        'size': {
                            'height': 256,
                            'width': 128,
                        },
                        'metadata/path': tf.compat.as_bytes('path/to/xyz.jpg'),
                    },
                },
            ),
        ],
    )

  @parameterized.parameters(
      (
          {
              'integer': np.int32,
              'string': np.object_,
          },
      ),
      (
          {
              'integer': np.int32,
              'string': np.str_,
          },
      ),
      (
          {
              'integer': tf.int32,
              'string': tf.string,
          },
      ),
  )
  def test_feature_getitem(self, features_dict):
    fdict = features_lib.FeaturesDict(features_dict)
    self.assertEqual(fdict['integer'].dtype, np.int32)
    self.assertEqual(fdict['string'].dtype, np.object_)

  def test_feature__repr__(self):
    label = features_lib.ClassLabel(names=['m', 'f'])
    feature_dict = features_lib.FeaturesDict({
        'metadata': features_lib.Sequence(
            {
                'frame': features_lib.Image(shape=(32, 32, 3)),
            }
        ),
        'label': features_lib.Sequence(label),
    })

    self.assertEqual(
        repr(feature_dict),
        textwrap.dedent(
            """\
        FeaturesDict({
            'label': Sequence(ClassLabel(shape=(), dtype=int64, num_classes=2)),
            'metadata': Sequence({
                'frame': Image(shape=(32, 32, 3), dtype=uint8),
            }),
        })"""
        ),
    )

  def test_feature_save_load_metadata_slashes(self):
    with testing.tmp_dir() as data_dir:
      fd = features_lib.FeaturesDict({
          'image/frame': features_lib.Image(shape=(32, 32, 3)),
          'image/label': features_lib.ClassLabel(num_classes=2),
      })
      fd.save_metadata(data_dir)
      fd.load_metadata(data_dir)

  @parameterized.parameters((np.int32), (tf.int32))
  def test_repr_tensor(self, dtype):
    # Top level Tensor is printed expanded
    self.assertEqual(
        repr(features_lib.Tensor(shape=(), dtype=tf.int32)),
        'Tensor(shape=(), dtype=int32)',
    )

    # Sequences colapse tensor repr
    self.assertEqual(
        repr(features_lib.Sequence(tf.int32)),
        'Sequence(int32)',
    )

    class ChildTensor(features_lib.Tensor):
      pass

    self.assertEqual(
        repr(
            features_lib.FeaturesDict({
                'colapsed': features_lib.Tensor(shape=(), dtype=dtype),
                # Tensor with defined shape are printed expanded
                'noncolapsed': features_lib.Tensor(shape=(1,), dtype=dtype),
                # Tensor inherited are expanded
                'child': ChildTensor(shape=(), dtype=dtype),
            })
        ),
        textwrap.dedent(
            """\
        FeaturesDict({
            'child': ChildTensor(shape=(), dtype=int32),
            'colapsed': int32,
            'noncolapsed': Tensor(shape=(1,), dtype=int32),
        })"""
        ),
    )


def test_custom_feature_connector(tmp_path: pathlib.Path):
  module_name = 'tensorflow_datasets.core.features.test_feature'
  feature_qualname = f'{module_name}.CustomFeatureConnector'
  assert module_name not in sys.modules
  assert (
      feature_qualname not in features_lib.FeatureConnector._registered_features
  )

  # Save tfds.feature.Tensor to avoid importing the custom feature connector
  feature = features_lib.Tensor(shape=(), dtype=np.int32)
  feature.save_config(tmp_path)

  # Update the features.json file to simulate as if the original
  # FeatureConnector had been save
  feature_path = tmp_path / 'features.json'
  content = feature_path.read_text()
  content = content.replace(
      'tensorflow_datasets.core.features.tensor_feature.Tensor',
      feature_qualname,
  )
  feature_path.write_text(content)

  # Loading will load the custom feature connector, even if it is not
  # registered yet.
  feature = features_lib.FeatureConnector.from_config(tmp_path)

  # After loading, the custom feature connector is registered
  assert module_name in sys.modules
  assert feature_qualname in features_lib.FeatureConnector._registered_features

  # Check that the loaded feature is the custom feature
  from tensorflow_datasets.core.features import test_feature  # pylint: disable=g-import-not-at-top

  assert isinstance(feature, test_feature.CustomFeatureConnector)


def test_tensor_feature_backward_compatibility():
  registered = features_lib.FeatureConnector._registered_features
  cls = features_lib.Tensor
  module_base = 'tensorflow_datasets.core.features.'
  # Both aliases are registered
  assert registered[module_base + 'tensor_feature.Tensor'] is cls
  assert registered[module_base + 'feature.Tensor'] is cls


def test_from_json_content_backward_compatibility():
  legacy_json = {
      'type': 'tensorflow_datasets.core.features.features_dict.FeaturesDict',
      'content': {
          'input': {
              'type': 'tensorflow_datasets.core.features.image_feature.Image',
              'content': {
                  'shape': [28, 28, 3],
                  'dtype': 'uint8',
                  'encoding_format': 'png',
              },
          },
          'target': {
              'type': 'tensorflow_datasets.core.features.class_label_feature.ClassLabel',
              'content': {'num_classes': 10},
          },
      },
  }
  parsed_features_dict = features_lib.FeatureConnector.from_json(legacy_json)
  assert parsed_features_dict.keys() == set(['input', 'target'])


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defined Reader and ReadInstruction to read tfrecord files."""

from __future__ import annotations

import functools
import os
import re
from typing import Any, Callable, List, NamedTuple, Optional, Sequence

from absl import logging
import numpy as np
from tensorflow_datasets.core import example_parser
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core import tf_compat
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.utils import file_utils
from tensorflow_datasets.core.utils import read_config as read_config_lib
from tensorflow_datasets.core.utils import shard_utils
from tensorflow_datasets.core.utils import tree_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

Tree = utils.Tree
TreeDict = utils.TreeDict
Tensor = utils.Tensor

ParseFn = Callable[[Tensor], TreeDict[Tensor]]
DecodeFn = Callable[[TreeDict[Tensor]], TreeDict[Tensor]]


# Use NamedTuple, as it is preserved by `tf.data.Dataset`
class _IdExample(NamedTuple):
  id: Any  # tf.string
  example: Any  # TreeDict[tf.any]


class _Instruction(NamedTuple):
  filepath: tf.string  # e.g. '/path/to/../train.tfrecord'
  filename: tf.string  # e.g. 'train.tfrecord'
  tfds_id_prefix: tf.string  # e.g. 'train.tfrecord' or 'folder1/train.tfrecord'
  skip: tf.int64
  take: tf.int64


def _get_dataset_from_filename(
    instruction: _Instruction,
    do_skip: bool,
    do_take: bool,
    file_format: file_adapters.FileFormat,
    add_tfds_id: bool,
    override_buffer_size: Optional[int] = None,
) -> tf.data.Dataset:
  """Returns a tf.data.Dataset instance from given instructions."""
  ds = file_adapters.ADAPTER_FOR_FORMAT[file_format].make_tf_data(
      instruction.filepath, buffer_size=override_buffer_size
  )
  if do_skip:
    ds = ds.skip(instruction.skip)
  if do_take:
    ds = ds.take(instruction.take)
  if add_tfds_id:  # For each example, generate a unique id.
    id_ds = _make_id_dataset(
        filename=instruction.tfds_id_prefix,
        start_index=instruction.skip if do_skip else 0,
    )
    ds = tf.data.Dataset.zip(_IdExample(id=id_ds, example=ds))
  return ds


def _make_id(
    filename: tf.Tensor,  # tf.Tensor[tf.string]
    index: tf.Tensor,  # tf.Tensor[tf.int64]
) -> tf.Tensor:  # tf.Tensor[tf.int64]
  """Creates the unique example ID."""
  return tf.strings.join(
      [filename, tf.strings.as_string(index)],
      separator='__',
  )


def _make_id_dataset(
    filename: tf.Tensor,  # tf.Tensor[tf.string]
    start_index: tf.Tensor,  # tf.Tensor[tf.int64]
) -> tf.data.Dataset:
  """Creates the dataset generating unique example IDs."""
  ds = tf.data.experimental.Counter(start_index)
  ds = ds.map(lambda index: _make_id(filename=filename, index=index))
  return ds


def _decode_with_id(
    id_ex: _IdExample,
    decode_fn: DecodeFn,
) -> TreeDict[Tensor]:
  """Add the `'tfds_id': ...` to the decoded example `dict`."""
  if not isinstance(id_ex, _IdExample):
    raise TypeError(
        f'Reader should return _IdExample when `add_tfds_id=True`. Got: {id_ex}'
    )
  decoded_ex = decode_fn(id_ex.example)
  if not isinstance(decoded_ex, dict):
    raise TypeError(
        f'Features should be `dict` when `add_tfds_id=True`. Got: {decoded_ex}'
    )
  decoded_ex['tfds_id'] = id_ex.id
  return decoded_ex


def _get_tfds_id_prefixes(
    file_instructions: Sequence[shard_utils.FileInstruction],
) -> dict[str, str]:
  """Returns the tfds id prefix per filename.

  If the file instructions are for files in different directories, then the TFDS
  id prefix includes part of the folder. This is to avoid duplicate TFDS ids
  when different directories contain the same filenames. For example, if the
  file instructions contain file paths `['/a/b/c', '/x/y/c']`, then this returns
  `{'/a/b/c': 'b/c', '/x/y/c': 'y/c'}`.

  Arguments:
    file_instructions: the file instructions for which to get TFDS id prefixes.

  Returns:
    the TFDS id prefix per filename.
  """
  dirnames = set(i.dirname() for i in file_instructions)
  reversed_dirnames = [d[::-1] for d in dirnames]
  common_suffix = os.path.commonprefix(reversed_dirnames)[::-1]

  def get_tfds_id(instruction: shard_utils.FileInstruction) -> str:
    # Return the file name when all files are in the same directory.
    if len(dirnames) == 1:
      return instruction.basename()
    # Otherwise, return the relative path starting from the deepest folder that
    # differs.
    return re.sub(
        rf'.*\/([^/]+{common_suffix}/[^/]+)', r'\1', instruction.filename
    )

  return {fi.filename: get_tfds_id(fi) for fi in file_instructions}


def _read_files(
    file_instructions: Sequence[shard_utils.FileInstruction],
    read_config: read_config_lib.ReadConfig,
    shuffle_files: bool,
    disable_shuffling: bool,
    file_format: file_adapters.FileFormat,
) -> tf.data.Dataset:
  """Returns tf.data.Dataset for given file instructions.

  Args:
    file_instructions: the information on the files to read including
      `ds.skip().take()`
    read_config: Additional options to configure the input pipeline (e.g. seed,
      num parallel reads,...).
    shuffle_files: Defaults to False. True to shuffle input files.
    disable_shuffling: Specifies if the dataset being read has shuffling
      disabled.
    file_format: Format of the record files in which the dataset will be
      read/written from.

  Returns:
    The dataset object.
  """

  def assert_cardinality_and_apply_options(ds):
    # If the number of examples read in the tf-record is known, we forward
    # the information to the tf.data.Dataset object.
    # Check the `tf.data.experimental` for backward compatibility with TF <= 2.1
    if (
        read_config.assert_cardinality
        and not read_config.input_context
        and hasattr(  # TODO(epot): Restore cardinality
            tf.data.experimental, 'assert_cardinality'
        )
    ):
      # TODO(b/154963426): Replace by per-shard cardinality (warning if
      # `experimental_interleave_sort_fn` is set).
      cardinality = sum(f.take for f in file_instructions)
      ds = ds.apply(tf.data.experimental.assert_cardinality(cardinality))

    return ds.with_options(read_config.options)  # Additional users options

  def validate_input_context():
    if not read_config.input_context:
      raise ValueError(
          'Cannot shard the pipeline with undefined `input_context`.'
      )
    try:
      num_input_pipelines = int(read_config.input_context.num_input_pipelines)
    except TypeError as e:
      logging.warning('Unable to validate `read_config.input_context`: %s', e)
      return
    if num_input_pipelines > 1 and len(file_instructions) < num_input_pipelines:
      raise ValueError(
          'Cannot shard the pipeline with given `input_context`.'
          '`num_shards={}` but `num_input_pipelines={}`. This means that some '
          "workers won't read any data. To shard the data, you may want to "
          'use the subsplit API instead: '
          'https://www.tensorflow.org/datasets/splits'.format(
              len(file_instructions),
              read_config.input_context.num_input_pipelines,
          )
      )

  # Eventually apply a transformation to the instruction function.
  # This allow the user to have direct control over the interleave order.
  if read_config.experimental_interleave_sort_fn is not None:
    file_instructions = read_config.experimental_interleave_sort_fn(
        file_instructions
    )

  do_skip = any(f.skip > 0 for f in file_instructions)
  do_take = any(not f.takes_all for f in file_instructions)

  tfds_id_prefixes = _get_tfds_id_prefixes(file_instructions)

  # Transpose the list[dict] into dict[list]
  tensor_inputs = _Instruction(
      filename=[os.path.basename(i.filename) for i in file_instructions],
      filepath=[os.fspath(i.filename) for i in file_instructions],
      tfds_id_prefix=[tfds_id_prefixes[i.filename] for i in file_instructions],
      # skip/take need to be converted to int64 explicitly
      skip=np.array([i.skip for i in file_instructions], dtype=np.int64),
      take=np.array([i.take for i in file_instructions], dtype=np.int64),
  )

  cycle_length = read_config.interleave_cycle_length
  block_length = read_config.interleave_block_length

  if cycle_length == read_config_lib.MISSING:
    cycle_length = _get_default_interleave_cycle_length(
        disable_shuffling=disable_shuffling
    )

  if disable_shuffling:
    _verify_read_config_for_ordered_dataset(
        read_config,
        interleave_cycle_length=cycle_length,
        shuffle_files=shuffle_files,
    )


  unique_folders = set(os.path.dirname(path) for path in tensor_inputs.filepath)
  logging.info(
      'Creating a tf.data.Dataset reading %d files located in folders: %s.',
      len(tensor_inputs.filepath),
      ', '.join(sorted(unique_folders)),
  )

  instruction_ds = tf.data.Dataset.from_tensor_slices(tensor_inputs)

  # On distributed environments, we can shard per-file if a
  # `tf.distribute.InputContext` object is provided (e.g. from
  # `experimental_distribute_datasets_from_function`)
  if read_config.input_context:
    validate_input_context()
    instruction_ds = instruction_ds.shard(
        num_shards=read_config.input_context.num_input_pipelines,
        index=read_config.input_context.input_pipeline_id,
    )

  # If shuffle is True, we shuffle the instructions/shards
  if shuffle_files:
    instruction_ds = instruction_ds.shuffle(
        len(file_instructions),
        seed=read_config.shuffle_seed,
        reshuffle_each_iteration=read_config.shuffle_reshuffle_each_iteration,
    )

  if read_config.repeat_filenames:
    instruction_ds = instruction_ds.repeat()

  deterministic = True
  # If shuffling is true and the seed is not set, then the pipeline is allowed
  # to be non-deterministic. Note that setting the deterministic option in
  # tf.data.Options() sets this option globally. This means that also nested
  # fields will appear in undeterministic order.
  if (
      shuffle_files
      and read_config.shuffle_seed is None
      and tf_compat.get_option_deterministic(read_config.options) is None
  ):
    deterministic = False

  ds = instruction_ds.interleave(
      functools.partial(
          _get_dataset_from_filename,
          do_skip=do_skip,
          do_take=do_take,
          file_format=file_format,
          add_tfds_id=read_config.add_tfds_id,
          override_buffer_size=read_config.override_buffer_size,
      ),
      cycle_length=cycle_length,
      block_length=block_length,
      num_parallel_calls=read_config.num_parallel_calls_for_interleave_files,
      deterministic=deterministic,
  )

  return assert_cardinality_and_apply_options(ds)


def _get_default_interleave_cycle_length(disable_shuffling: bool) -> int:
  if disable_shuffling:
    logging.info(
        '`interleave_cycle_length` set to 1 to read examples in order.'
    )
    return 1
  else:
    return 16


def _verify_read_config_for_ordered_dataset(
    read_config: read_config_lib.ReadConfig,
    interleave_cycle_length: int,
    shuffle_files: bool,
) -> None:
  """Check that read parameters will not affect the ordering of the dataset.

  The user can bypass the error by setting `enable_ordering_guard=False`.

  Args:
    read_config: The input pipeline options.
    interleave_cycle_length: Cycle length for `tf.data.Dataset.interleave`.
    shuffle_files: If True, input files are shuffled before being read.
  """
  error_messages = []
  if shuffle_files:
    error_messages.append(
        "Dataset is an ordered dataset ('disable_shuffling=True'), but examples"
        ' will not be read in order because `shuffle_files=True`.'
    )
  if interleave_cycle_length != 1:
    error_messages.append(
        "Dataset is an ordered dataset ('disable_shuffling=True'), but examples"
        ' will not be read in order because `ReadConfig.interleave_cycle_length'
        ' != 1`.'
    )
  if error_messages:
    error_message = '\n'.join(error_messages)
    if read_config.enable_ordering_guard:
      raise ValueError(error_message)
    else:
      logging.warning(error_message)


class Reader(object):
  """Build a tf.data.Dataset object out of Instruction instance(s).

  This class should not typically be exposed to the TFDS user.
  """

  def __init__(
      self,
      path,  # TODO(b/216427814) remove this as it isn't used anymore
      example_specs,
      file_format=file_adapters.DEFAULT_FILE_FORMAT,
  ):
    """Initializes Reader.

    Args:
      path (str): path where tfrecords are stored.
      example_specs: spec to build ExampleParser.
      file_format: file_adapters.FileFormat, format of the record files in which
        the dataset will be read/written from.
    """
    self._path = path
    self._parser = example_parser.ExampleParser(example_specs)
    self._file_format = file_format

  def read(
      self,
      *,
      instructions: Tree[splits_lib.SplitArg],
      split_infos: List[splits_lib.SplitInfo],
      read_config: read_config_lib.ReadConfig,
      shuffle_files: bool,
      disable_shuffling: bool = False,
      decode_fn: Optional[DecodeFn] = None,
  ) -> Tree[tf.data.Dataset]:
    """Returns tf.data.Dataset instance(s).

    Args:
      instructions (ReadInstruction, List[], Dict[]): instruction(s) to read.
        Instructions can be string and will then be passed to the Instruction
        constructor as it.
      split_infos: the available splits for dataset.
      read_config: `tfds.ReadConfig`, the input pipeline options
      shuffle_files (bool): If True, input files are shuffled before being read.
      disable_shuffling: Specifies if the dataset being read has shuffling
        disabled.
      decode_fn: Eventual additional processing to apply to the example after
        deserialization.

    Returns:
       a single tf.data.Dataset instance if instruction is a single
       ReadInstruction instance. Otherwise a dict/list of tf.data.Dataset
       corresponding to given instructions param shape.
    """

    splits_dict = splits_lib.SplitDict(split_infos=split_infos)

    def _read_instruction_to_ds(instruction):
      file_instructions = splits_dict[instruction].file_instructions
      return self.read_files(
          file_instructions,
          read_config=read_config,
          shuffle_files=shuffle_files,
          disable_shuffling=disable_shuffling,
          decode_fn=decode_fn,
      )

    return tree_utils.map_structure(_read_instruction_to_ds, instructions)

  def read_files(
      self,
      file_instructions: Sequence[shard_utils.FileInstruction],
      *,
      read_config: read_config_lib.ReadConfig,
      shuffle_files: bool,
      disable_shuffling: bool = False,
      decode_fn: Optional[DecodeFn] = None,
  ) -> tf.data.Dataset:
    """Returns single tf.data.Dataset instance for the set of file instructions.

    Args:
      file_instructions: The files information. The filenames contains the
        relative path, not absolute. skip/take indicates which example read in
        the shard: `ds.skip().take()`
      read_config: The input pipeline options
      shuffle_files: If True, input files are shuffled before being read.
      disable_shuffling: Specifies if the dataset being read has shuffling
        disabled.
      decode_fn: Eventual additional processing to apply to the example after
        deserialization.

    Returns:
       a tf.data.Dataset instance.
    """
    if not file_instructions:
      msg = f'Instruction {file_instructions} corresponds to no data!'
      raise ValueError(msg)

    # Read serialized example (eventually with `tfds_id`)
    ds = _read_files(
        file_instructions=file_instructions,
        read_config=read_config,
        shuffle_files=shuffle_files,
        disable_shuffling=disable_shuffling,
        file_format=self._file_format,
    )

    # Parse and decode
    def parse_and_decode(ex: Tensor) -> TreeDict[Tensor]:
      # TODO(pierrot): `parse_example` uses
      # `tf.io.parse_single_example`. It might be faster to use `parse_example`,
      # after batching.
      # https://www.tensorflow.org/api_docs/python/tf/io/parse_example
      ex = self._parser.parse_example(ex)
      if decode_fn:
        ex = decode_fn(ex)
      return ex

    # Eventually add the `tfds_id` after the decoding
    if read_config and read_config.add_tfds_id:
      parse_and_decode = functools.partial(
          _decode_with_id, decode_fn=parse_and_decode
      )

    ds = ds.map(
        parse_and_decode,
        num_parallel_calls=read_config.num_parallel_calls_for_decode,
    )
    return ds



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.split_builder."""

import collections
import io
import os
import pathlib

import apache_beam as beam
import psutil
import pytest
from tensorflow_datasets import testing
from tensorflow_datasets.core import split_builder as split_builder_lib


def _inc_placeholder_counter(x):
  beam.metrics.Metrics.counter('some_namespace', 'some_counter').inc()
  return x


@pytest.fixture(name='split_builder')
def fixture_split_builder():
  builder = testing.DummyMnist()
  return split_builder_lib.SplitBuilder(
      split_dict=builder.info.splits,
      features=builder.info.features,
      dataset_size=builder.info.dataset_size,
      beam_options=None,
      beam_runner=None,
      max_examples_per_split=None,
  )


def test_beam(
    tmp_path: pathlib.Path, split_builder: split_builder_lib.SplitBuilder
):
  """Test that `maybe_beam_pipeline` behave as `beam.Pipeline()`."""
  path = tmp_path / 'out.txt'
  with split_builder.maybe_beam_pipeline() as pipeline_proxy:
    ptransform = (
        beam.Create(range(9))
        | beam.Map(lambda x: x * 10)
        | beam.Map(_inc_placeholder_counter)
        | beam.io.WriteToText(os.fspath(path), shard_name_template='')
    )
    _ = split_builder.beam_pipeline | ptransform
  result = pipeline_proxy.result
  # counters = metrics.get_metrics(result, 'some_namespace').counters
  mfilter = beam.metrics.MetricsFilter().with_namespaces(['some_namespace'])
  all_metrics = result.metrics().query(mfilter)
  counters = all_metrics['counters']
  assert counters[0].key.metric.name == 'some_counter'
  assert counters[0].committed == 9
  assert path.read_text() == '\n'.join(str(x * 10) for x in range(9)) + '\n'


def test_memory_warning(
    capsys, monkeypatch, split_builder: split_builder_lib.SplitBuilder
):
  def mock_psutil_virtual_memory():
    MockData = collections.namedtuple('MockData', ['total'])
    return MockData(total=0)

  monkeypatch.setattr(psutil, 'virtual_memory', mock_psutil_virtual_memory)
  monkeypatch.setattr('sys.stdin', io.StringIO('n\n'))

  with pytest.raises(SystemExit) as pytest_wrapped_e:
    with split_builder.maybe_beam_pipeline():
      _ = split_builder.beam_pipeline

  assert pytest_wrapped_e.type == SystemExit
  assert pytest_wrapped_e.value.code == 1

  captured = capsys.readouterr()
  assert 'Continue?' in captured.out



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""clinc_oos dataset."""

from __future__ import annotations

import csv
import os

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

_DOWNLOAD_URL = 'https://github.com/jereliu/datasets/raw/master/clinc_oos.zip'

_CITATION = """
@inproceedings{larson-etal-2019-evaluation,
    title = "An Evaluation Dataset for Intent Classification and Out-of-Scope Prediction",
    author = "Larson, Stefan  and
      Mahendran, Anish  and
      Peper, Joseph J.  and
      Clarke, Christopher  and
      Lee, Andrew  and
      Hill, Parker  and
      Kummerfeld, Jonathan K.  and
      Leach, Kevin  and
      Laurenzano, Michael A.  and
      Tang, Lingjia  and
      Mars, Jason",
    booktitle = "Proceedings of the 2019 Conference on Empirical Methods in Natural Language Processing and the 9th International Joint Conference on Natural Language Processing (EMNLP-IJCNLP)",
    month = nov,
    year = "2019",
    address = "Hong Kong, China",
    publisher = "Association for Computational Linguistics",
    url = "https://www.aclweb.org/anthology/D19-1131",
    doi = "10.18653/v1/D19-1131",
    pages = "1311--1316",
}
"""

_DESCRIPTION = """
Task-oriented dialog systems need to know when a query falls outside their range of supported intents, but current text classification corpora only define label sets that cover every example. We introduce a new dataset that includes queries that are out-of-scope (OOS), i.e., queries that do not fall into any of the system's supported intents. This poses a new challenge because models cannot assume that every query at inference time belongs to a system-supported intent class. Our dataset also covers 150 intent classes over 10 domains, capturing the breadth that a production task-oriented agent must handle. It offers a way of more rigorously and realistically benchmarking text classification in task-driven dialog systems.
"""


class ClincOOS(tfds.core.GeneratorBasedBuilder):
  """CLINC Dataset for Intent Classification and Out-of-Scope (OOS) Detection.

  This dataset is for evaluating the performance of intent classification
  systems in the presence of "out-of-scope" queries. By "out-of-scope",
  we mean queries that do not fall into any of the system-supported intent
  classes. Most datasets include only data that is "in-scope".
  Our dataset includes both in-scope and out-of-scope data.

  This version of the CLINC OOS dataset contains 150 "in-scope" intents from
  10 domains. Each intent has 100 train, 20 validation, and 30 test samples.
  There are 100 train and validation out-of-scope samples, and 1000 out-of-scope
  test samples. It also contains labels for intent domain which are not included
  in the original dataset.
  """

  VERSION = tfds.core.Version('0.1.0')

  def _info(self):
    # TODO(clinc_oos): Specifies the tfds.core.DatasetInfo object
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            'text': tfds.features.Text(),
            'intent': np.int32,
            'domain': np.int32,
            'intent_name': tfds.features.Text(),
            'domain_name': tfds.features.Text(),
        }),
        supervised_keys=('text', 'intent'),
        homepage='https://github.com/clinc/oos-eval/',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    data_path = dl_manager.download_and_extract(_DOWNLOAD_URL)
    return [
        tfds.core.SplitGenerator(
            name='train',
            gen_kwargs={'filename': os.path.join(data_path, 'train.csv')},
        ),
        tfds.core.SplitGenerator(
            name='test',
            gen_kwargs={'filename': os.path.join(data_path, 'test.csv')},
        ),
        tfds.core.SplitGenerator(
            name='validation',
            gen_kwargs={'filename': os.path.join(data_path, 'val.csv')},
        ),
        tfds.core.SplitGenerator(
            name='train_oos',
            gen_kwargs={'filename': os.path.join(data_path, 'train_ood.csv')},
        ),
        tfds.core.SplitGenerator(
            name='test_oos',
            gen_kwargs={'filename': os.path.join(data_path, 'test_ood.csv')},
        ),
        tfds.core.SplitGenerator(
            name='validation_oos',
            gen_kwargs={'filename': os.path.join(data_path, 'val_ood.csv')},
        ),
    ]

  def _generate_examples(self, filename):
    """Yields examples."""
    with epath.Path(filename).open() as f:
      reader = csv.DictReader(f)
      for row_id, row in enumerate(reader):
        example = {}
        example['text'] = row['text']
        example['intent'] = row['intent']
        example['domain'] = row['domain']
        example['intent_name'] = row['intent_name']
        example['domain_name'] = row['domain_name']

        yield row_id, example
